#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2020 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides classes for various bytecode instrumentations."""
from __future__ import annotations

import json
import logging
from abc import abstractmethod
from types import CodeType
from typing import TYPE_CHECKING

from bytecode import BasicBlock, Bytecode, Compare, ControlFlowGraph, Instr

from pynguin.analyses.controlflow.cfg import CFG
from pynguin.analyses.controlflow.controldependencegraph import ControlDependenceGraph
from pynguin.analyses.seeding.constantseeding import DynamicConstantSeeding
from pynguin.testcase.execution import (
    CodeObjectMetaData,
    ExecutionTracer,
    PredicateMetaData,
)

if TYPE_CHECKING:
    from pynguin.analyses.controlflow.programgraph import ProgramGraphNode

CODE_OBJECT_ID_KEY = "code_object_id"


# pylint:disable=too-few-public-methods
class Instrumentation:
    """Abstract base class for bytecode instrumentations.

    General notes:

    When calling a method on an object, the arguments have to be on top of the stack.
    In most cases, we need to rotate the items on the stack with ROT_THREE or ROT_FOUR
    to reorder the elements accordingly.

    A POP_TOP instruction is required after calling a method, because each method
    implicitly returns None."""

    @abstractmethod
    def instrument_module(self, module_code: CodeType) -> CodeType:
        """Instrument the given code object of a module.

        Args:
            module_code: The code object of the module

        Returns:
            The instrumented code object of the module
        """

    @staticmethod
    def _create_consecutive_blocks(
        bytecode_cfg: ControlFlowGraph, first: BasicBlock, amount: int
    ) -> tuple[BasicBlock, ...]:
        """Split the given basic block into more blocks.

        The blocks are consecutive in the list of basic blocks.

        Args:
            bytecode_cfg: The control-flow graph
            first: The first basic block
            amount: The amount of consecutive blocks that should be created.

        Returns:
            A tuple of consecutive basic blocks
        """
        assert amount > 0, "Amount of created basic blocks must be positive."
        current: BasicBlock = first
        nodes: list[BasicBlock] = []
        # Can be any instruction, as it is discarded anyway.
        dummy_instruction = Instr("POP_TOP")
        for _ in range(amount):
            # Insert dummy instruction, which we can use to split off another block
            current.insert(0, dummy_instruction)
            current = bytecode_cfg.split_block(current, 1)
            nodes.append(current)

        # Move instructions back to first block.
        first.clear()
        first.extend(current)
        # Clear instructions in all created blocks.
        for node in nodes:
            node.clear()
        return tuple(nodes)


class BranchCoverageInstrumentation(Instrumentation):
    """Instruments code objects to enable tracking branch distances and thus
    branch coverage."""

    # As of CPython 3.8, there are a few compare ops for which we can't really
    # compute a sensible branch distance. So for now, we just ignore those
    # comparisons and just track their boolean value.
    # As of CPython 3.9, this is no longer a compare op but instead
    # a JUMP_IF_NOT_EXC_MATCH, which we also handle as boolean based jump.
    _IGNORED_COMPARE_OPS: set[Compare] = {Compare.EXC_MATCH}

    # Conditional jump operations are the last operation within a basic block
    _JUMP_OP_POS = -1

    # If a conditional jump is based on a comparison, it has to be the second-to-last
    # instruction within the basic block.
    _COMPARE_OP_POS = -2

    _logger = logging.getLogger(__name__)

    def __init__(self, tracer: ExecutionTracer) -> None:
        self._tracer = tracer

    def _instrument_inner_code_objects(
        self, code: CodeType, parent_code_object_id: int
    ) -> CodeType:
        """Apply the instrumentation to all constants of the given code object.

        Args:
            code: the Code Object that should be instrumented.
            parent_code_object_id: the id of the parent code object, if any.

        Returns:
            the code object whose constants were instrumented.
        """
        new_consts = []
        for const in code.co_consts:
            if isinstance(const, CodeType):
                # The const is an inner code object
                new_consts.append(
                    self._instrument_code_recursive(
                        const, parent_code_object_id=parent_code_object_id
                    )
                )
            else:
                new_consts.append(const)
        return code.replace(co_consts=tuple(new_consts))

    def _instrument_code_recursive(
        self,
        code: CodeType,
        parent_code_object_id: int | None = None,
    ) -> CodeType:
        """Instrument the given Code Object recursively.

        Args:
            code: The code object that should be instrumented
            parent_code_object_id: The ID of the optional parent code object

        Returns:
            The instrumented code object
        """
        self._logger.debug("Instrumenting Code Object for %s", code.co_name)
        cfg = CFG.from_bytecode(Bytecode.from_code(code))
        cdg = ControlDependenceGraph.compute(cfg)
        code_object_id = self._tracer.register_code_object(
            CodeObjectMetaData(
                code_object=code,
                parent_code_object_id=parent_code_object_id,
                cfg=cfg,
                cdg=cdg,
            )
        )
        # Overwrite/Set docstring to carry tagging information, i.e.,
        # the code object id. Convert to JSON string because I'm not sure where this
        # value might be used in CPython.
        cfg.bytecode_cfg().docstring = json.dumps({CODE_OBJECT_ID_KEY: code_object_id})
        assert cfg.entry_node is not None, "Entry node cannot be None."
        real_entry_node = cfg.get_successors(cfg.entry_node).pop()  # Only one exists!
        assert real_entry_node.basic_block is not None, "Basic block cannot be None."
        self._add_code_object_executed(real_entry_node.basic_block, code_object_id)
        self._instrument_cfg(cfg, code_object_id)
        return self._instrument_inner_code_objects(
            cfg.bytecode_cfg().to_code(), code_object_id
        )

    def _instrument_cfg(self, cfg: CFG, code_object_id: int) -> None:
        """Instrument the bytecode cfg associated with the given CFG.

        Args:
            cfg: The CFG that overlays the bytecode cfg.
            code_object_id: The id of the code object which contains this CFG.
        """
        # Required to transform for loops.
        for node in cfg.nodes:
            predicate_id = self._instrument_node(cfg, code_object_id, node)
            if predicate_id is not None:
                node.predicate_id = predicate_id

    def _instrument_node(
        self,
        cfg: CFG,
        code_object_id: int,
        node: ProgramGraphNode,
    ) -> int | None:
        """Instrument a single node in the CFG.

        Currently we only instrument conditional jumps and for loops.

        Args:
            cfg: The containing CFG.
            code_object_id: The containing Code Object
            node: The node that should be instrumented.

        Returns:
            A predicate id, if the contained a predicate which was instrumented.
        """
        predicate_id: int | None = None
        # Not every block has an associated basic block, e.g. the artificial exit node.
        if not node.is_artificial:
            assert (
                node.basic_block is not None
            ), "Non artificial node does not have a basic block."
            assert len(node.basic_block) > 0, "Empty basic block in CFG."
            maybe_jump: Instr = node.basic_block[self._JUMP_OP_POS]
            maybe_compare: Instr | None = (
                node.basic_block[self._COMPARE_OP_POS]
                if len(node.basic_block) > 1
                else None
            )
            if isinstance(maybe_jump, Instr):
                if maybe_jump.name == "FOR_ITER":
                    predicate_id = self._instrument_for_loop(cfg, node, code_object_id)
                elif maybe_jump.is_cond_jump():
                    predicate_id = self._instrument_cond_jump(
                        code_object_id,
                        maybe_compare,
                        maybe_jump,
                        node.basic_block,
                        node,
                    )
        return predicate_id

    def _instrument_cond_jump(
        self,
        code_object_id: int,
        maybe_compare: Instr | None,
        jump: Instr,
        block: BasicBlock,
        node: ProgramGraphNode,
    ) -> int:
        # pylint:disable=too-many-arguments
        """Instrument a conditional jump.

        If it is based on a prior comparison, we track
        the compared values, otherwise we just track the truthiness of the value on top
        of the stack.

        Args:
            code_object_id: The id of the containing Code Object.
            maybe_compare: The comparison operation, if any.
            jump: The jump operation.
            block: The containing basic block.
            node: The associated node from the CFG.

        Returns:
            The id that was assigned to the predicate.
        """
        if (
            maybe_compare is not None
            and isinstance(maybe_compare, Instr)
            and (
                (
                    maybe_compare.name == "COMPARE_OP"
                    and maybe_compare.arg not in self._IGNORED_COMPARE_OPS
                )
                or maybe_compare.name in ("IS_OP", "CONTAINS_OP")
            )
        ):
            return self._instrument_compare_based_conditional_jump(
                block, code_object_id, node
            )
        # Up to 3.9, there was COMPARE_OP EXC_MATCH which was handled below
        # Beginning with 3.9, there is a combined compare+jump op, which is handled
        # here.
        if jump.name == "JUMP_IF_NOT_EXC_MATCH":
            return self._instrument_exception_based_conditional_jump(
                block, code_object_id, node
            )
        return self._instrument_bool_based_conditional_jump(block, code_object_id, node)

    def _instrument_bool_based_conditional_jump(
        self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
    ) -> int:
        """Instrument boolean-based conditional jumps.

        We add a call to the tracer which reports the value on which the conditional
        jump will be based.

        Args:
            block: The containing basic block.
            code_object_id: The id of the containing Code Object.
            node: The associated node from the CFG.

        Returns:
            The id assigned to the predicate.
        """
        lineno = block[self._JUMP_OP_POS].lineno
        predicate_id = self._tracer.register_predicate(
            PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
        )
        # Insert instructions right before the conditional jump.
        # We duplicate the value on top of the stack and report
        # it to the tracer.
        block[self._JUMP_OP_POS : self._JUMP_OP_POS] = [
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", self._tracer, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                ExecutionTracer.executed_bool_predicate.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("LOAD_CONST", predicate_id, lineno=lineno),
            Instr("CALL_METHOD", 2, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        return predicate_id

    def _instrument_compare_based_conditional_jump(
        self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
    ) -> int:
        """Instrument compare-based conditional jumps.

        We add a call to the tracer which reports the values that will be used
        in the following comparison operation on which the conditional jump is based.

        Args:
            block: The containing basic block.
            code_object_id: The id of the containing Code Object.
            node: The associated node from the CFG.

        Raises:
            RuntimeError: If an unknown operation is encountered.

        Returns:
            The id assigned to the predicate.
        """
        lineno = block[self._JUMP_OP_POS].lineno
        predicate_id = self._tracer.register_predicate(
            PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
        )
        operation = block[self._COMPARE_OP_POS]

        if operation.name == "COMPARE_OP":
            compare = operation.arg
        elif operation.name == "IS_OP":
            # Beginning with 3.9, there are separate OPs for various comparisons.
            compare = Compare.IS_NOT if operation.arg else Compare.IS
        elif operation.name == "CONTAINS_OP":
            compare = Compare.NOT_IN if operation.arg else Compare.IN
        else:
            raise RuntimeError(f"Unknown comparison OP {operation}")

        # Insert instructions right before the comparison.
        # We duplicate the values on top of the stack and report
        # them to the tracer.
        block[self._COMPARE_OP_POS : self._COMPARE_OP_POS] = [
            Instr("DUP_TOP_TWO", lineno=lineno),
            Instr("LOAD_CONST", self._tracer, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                ExecutionTracer.executed_compare_predicate.__name__,
                lineno=lineno,
            ),
            Instr("ROT_FOUR", lineno=lineno),
            Instr("ROT_FOUR", lineno=lineno),
            Instr("LOAD_CONST", predicate_id, lineno=lineno),
            Instr("LOAD_CONST", compare, lineno=lineno),
            Instr("CALL_METHOD", 4, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        return predicate_id

    def _instrument_exception_based_conditional_jump(
        self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
    ) -> int:
        """Instrument exception-based conditional jumps.

        We add a call to the tracer which reports the values that will be used
        in the following exception matching case.

        Args:
            block: The containing basic block.
            code_object_id: The id of the containing Code Object.
            node: The associated node from the CFG.

        Returns:
            The id assigned to the predicate.
        """
        lineno = block[self._JUMP_OP_POS].lineno
        predicate_id = self._tracer.register_predicate(
            PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
        )
        # Insert instructions right before the conditional jump.
        # We duplicate the values on top of the stack and report
        # them to the tracer.
        block[self._JUMP_OP_POS : self._JUMP_OP_POS] = [
            Instr("DUP_TOP_TWO", lineno=lineno),
            Instr("LOAD_CONST", self._tracer, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                ExecutionTracer.executed_exception_match.__name__,
                lineno=lineno,
            ),
            Instr("ROT_FOUR", lineno=lineno),
            Instr("ROT_FOUR", lineno=lineno),
            Instr("LOAD_CONST", predicate_id, lineno=lineno),
            Instr("CALL_METHOD", 3, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        return predicate_id

    def _add_code_object_executed(self, block: BasicBlock, code_object_id: int) -> None:
        """Add instructions at the beginning of the given basic block which inform
        the tracer, that the code object with the given id has been entered.

        Args:
            block: The entry basic block of a code object, i.e. the first basic block.
            code_object_id: The id that the tracer has assigned to the code object
                which contains the given basic block.
        """
        # Use line number of first instruction
        lineno = block[0].lineno
        # Insert instructions at the beginning.
        block[0:0] = [
            Instr("LOAD_CONST", self._tracer, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                ExecutionTracer.executed_code_object.__name__,
                lineno=lineno,
            ),
            Instr("LOAD_CONST", code_object_id, lineno=lineno),
            Instr("CALL_METHOD", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]

    def instrument_module(self, module_code: CodeType) -> CodeType:
        for const in module_code.co_consts:
            if isinstance(const, ExecutionTracer):
                # Abort instrumentation, since we have already
                # instrumented this code object.
                assert False, "Tried to instrument already instrumented module."
        return self._instrument_code_recursive(module_code)

    def _instrument_for_loop(
        self,
        cfg: CFG,
        node: ProgramGraphNode,
        code_object_id: int,
    ) -> int:
        """Transform the for loop whose header is defined in the given node.
        We only transform the underlying bytecode cfg, by partially unrolling the first
        iteration. For this, we add two basic blocks after the loop header:

        The first block is called, if the iterator on which the loop is based
        yields at least one element, in which case we report the boolean value True
        to the tracer, leave the yielded value of the iterator on top of the stack and
        jump to the the regular body of the loop.

        The second block is called, if the iterator on which the loop is based
        does not yield an element, in which case we report the boolean value False
        to the tracer and jump to the exit instruction of the loop.

        The original loop header is changed such that it either falls through to the
        first block or jumps to the second, if no element is yielded.

        Since Python is a structured programming language, there can be no jumps
        directly into the loop that bypass the loop header (e.g., GOTO).
        Jumps which reach the loop header from outside the loop will still target
        the original loop header, so they don't need to be modified.

        Args:
            cfg: The CFG that contains the loop
            node: The node which contains the header of the for loop.
            code_object_id: The id of the containing Code Object.

        Returns:
            The ID of the instrumented predicate
        """
        assert node.basic_block is not None, "Basic block of for loop cannot be None."
        for_instr = node.basic_block[self._JUMP_OP_POS]
        assert for_instr.name == "FOR_ITER"
        lineno = for_instr.lineno
        predicate_id = self._tracer.register_predicate(
            PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
        )
        for_loop_exit = for_instr.arg
        for_loop_body = node.basic_block.next_block

        # pylint:disable=unbalanced-tuple-unpacking
        entered, not_entered = self._create_consecutive_blocks(
            cfg.bytecode_cfg(), node.basic_block, 2
        )
        for_instr.arg = not_entered

        entered.extend(
            [
                Instr("LOAD_CONST", self._tracer, lineno=lineno),
                Instr(
                    "LOAD_METHOD",
                    ExecutionTracer.executed_bool_predicate.__name__,
                    lineno=lineno,
                ),
                Instr("LOAD_CONST", True, lineno=lineno),
                Instr("LOAD_CONST", predicate_id, lineno=lineno),
                Instr("CALL_METHOD", 2, lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("JUMP_ABSOLUTE", for_loop_body, lineno=lineno),
            ]
        )

        not_entered.extend(
            [
                Instr("LOAD_CONST", self._tracer, lineno=lineno),
                Instr(
                    "LOAD_METHOD",
                    ExecutionTracer.executed_bool_predicate.__name__,
                    lineno=lineno,
                ),
                Instr("LOAD_CONST", False, lineno=lineno),
                Instr("LOAD_CONST", predicate_id, lineno=lineno),
                Instr("CALL_METHOD", 2, lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("JUMP_ABSOLUTE", for_loop_exit, lineno=lineno),
            ]
        )

        return predicate_id


# pylint:disable=too-few-public-methods
class DynamicSeedingInstrumentation(Instrumentation):
    """Instruments code objects to enable dynamic constant seeding.

    Supported is collecting values of the types int, float and string.

    Instrumented are the common compare operations (==, !=, <, >, <=, >=) and the string
    methods contained in the STRING_FUNCTION_NAMES list. This means, if one of the
    above operations and methods is used in an if-conditional, corresponding values
    are added to the dynamic constant pool.

    The dynamic pool is implemented in the module constantseeding.py. The dynamicseeding
    module containes methods for managing the dynamic pool during the algorithm
    execution."""

    # Compare operations are only followed by one jump operation, hence they are on the
    # second to last position of the block.
    _COMPARE_OP_POS = -2

    #  If one of the considered string functions needing no argument is used in the if
    #  statement, it will be loaded in the third last position. After it comes the
    #  call of the method and the jump operation.
    _STRING_FUNC_POS = -3

    # If one of the considered string functions needing one argument is used in the if
    # statement, it will be loaded in the fourth last position. After it comes the
    # load of the argument, the call of the method and the jump
    # operation.
    _STRING_FUNC_POS_WITH_ARG = -4

    # A list containing the names of all string functions which are instrumented.
    _STRING_FUNCTION_NAMES = [
        "startswith",
        "endswith",
        "isalnum",
        "isalpha",
        "isdecimal",
        "isdigit",
        "isidentifier",
        "islower",
        "isnumeric",
        "isprintable",
        "isspace",
        "istitle",
        "isupper",
    ]

    _logger = logging.getLogger(__name__)

    def __init__(self, dynamic_constant_seeding: DynamicConstantSeeding):
        self._dynamic_constant_seeding = dynamic_constant_seeding

    def _instrument_startswith_function(self, block: BasicBlock) -> None:
        """Instruments the startswith function in bytecode. Stores for the expression
          'string1.startswith(string2)' the
           value 'string2 + string1' in the _dynamic_pool.

        Args:
            block: The basic block where the new instructions are inserted.
        """
        insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
        lineno = block[insert_pos].lineno
        block[insert_pos:insert_pos] = [
            Instr("DUP_TOP_TWO", lineno=lineno),
            Instr("ROT_TWO", lineno=lineno),
            Instr("BINARY_ADD", lineno=lineno),
            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                self._dynamic_constant_seeding.add_value.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("CALL_METHOD", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        self._logger.info("Instrumented startswith function")

    def _instrument_endswith_function(self, block: BasicBlock) -> None:
        """Instruments the endswith function in bytecode. Stores for the expression
         'string1.startswith(string2)' the
           value 'string1 + string2' in the _dynamic_pool.

        Args:
            block: The basic block where the new instructions are inserted.
        """
        insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
        lineno = block[insert_pos].lineno
        block[insert_pos:insert_pos] = [
            Instr("DUP_TOP_TWO", lineno=lineno),
            Instr("BINARY_ADD", lineno=lineno),
            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                DynamicConstantSeeding.add_value.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("CALL_METHOD", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        self._logger.info("Instrumented endswith function")

    def _instrument_string_function_without_arg(
        self, block: BasicBlock, function_name: str
    ) -> None:
        """Instruments the isalnum function in bytecode.

        Args:
            block: The basic block where the new instructions are inserted.
            function_name: The name of the function
        """
        insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
        lineno = block[insert_pos].lineno
        block[insert_pos:insert_pos] = [
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                DynamicConstantSeeding.add_value_for_strings.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("LOAD_CONST", function_name, lineno=lineno),
            Instr("CALL_METHOD", 2, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        self._logger.info("Instrumented string function")

    def _instrument_string_func(self, block: BasicBlock, function_name: str) -> None:
        """Calls the corresponding instrumentation method for the given function_name.

        Args:
            block: The block to instrument.
            function_name: The name of the function for which the method will be called.

        """
        if function_name == "startswith":
            self._instrument_startswith_function(block)
        elif function_name == "endswith":
            self._instrument_endswith_function(block)
        else:
            self._instrument_string_function_without_arg(block, function_name)

    def _instrument_compare_op(self, block: BasicBlock) -> None:
        """Instruments the compare operations in bytecode. Stores the values extracted
         at runtime.

        Args:
            block: The containing basic block.
        """
        lineno = block[self._COMPARE_OP_POS].lineno
        block[self._COMPARE_OP_POS : self._COMPARE_OP_POS] = [
            Instr("DUP_TOP_TWO", lineno=lineno),
            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                DynamicConstantSeeding.add_value.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("CALL_METHOD", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
            Instr(
                "LOAD_METHOD",
                DynamicConstantSeeding.add_value.__name__,
                lineno=lineno,
            ),
            Instr("ROT_THREE", lineno=lineno),
            Instr("ROT_THREE", lineno=lineno),
            Instr("CALL_METHOD", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
        self._logger.debug("Instrumented compare_op")

    def _instrument_inner_code_objects(self, code: CodeType) -> CodeType:
        """Apply the instrumentation to all constants of the given code object.

        Args:
            code: the Code Object that should be instrumented.

        Returns:
            the code object whose constants were instrumented.
        """
        new_consts = []
        for const in code.co_consts:
            if isinstance(const, CodeType):
                # The const is an inner code object
                new_consts.append(self._instrument_code_recursive(const))
            else:
                new_consts.append(const)
        return code.replace(co_consts=tuple(new_consts))

    def _instrument_code_recursive(
        self,
        code: CodeType,
    ) -> CodeType:
        """Instrument the given Code Object recursively.

        Args:
            code: The code object that should be instrumented

        Returns:
            The instrumented code object
        """
        self._logger.debug(
            "Instrumenting Code Object for dynamic seeding for %s", code.co_name
        )
        cfg = CFG.from_bytecode(Bytecode.from_code(code))

        assert cfg.entry_node is not None, "Entry node cannot be None."
        real_entry_node = cfg.get_successors(cfg.entry_node).pop()  # Only one exists!
        assert real_entry_node.basic_block is not None, "Basic block cannot be None."

        self._instrument_cfg(cfg)
        return self._instrument_inner_code_objects(cfg.bytecode_cfg().to_code())

    def _instrument_cfg(self, cfg: CFG) -> None:
        """Instrument the bytecode cfg associated with the given CFG.

        Args:
            cfg: The CFG that overlays the bytecode cfg.
        """
        # Attributes which store the predicate ids assigned to instrumented nodes.
        for node in cfg.nodes:
            self._instrument_node(node)

    def _instrument_node(
        self,
        node: ProgramGraphNode,
    ) -> None:
        """Instrument a single node in the CFG.

        Currently we only instrument conditional jumps and for loops.

        Args:
            node: The node that should be instrumented.
        """
        # Not every block has an associated basic block, e.g. the artificial exit node.
        if not node.is_artificial:
            assert (
                node.basic_block is not None
            ), "Non artificial node does not have a basic block."
            assert len(node.basic_block) > 0, "Empty basic block in CFG."
            maybe_compare: Instr | None = (
                node.basic_block[self._COMPARE_OP_POS]
                if len(node.basic_block) > 1
                else None
            )
            maybe_string_func: Instr | None = (
                node.basic_block[self._STRING_FUNC_POS]
                if len(node.basic_block) > 2
                else None
            )
            maybe_string_func_with_arg: Instr | None = (
                node.basic_block[self._STRING_FUNC_POS_WITH_ARG]
                if len(node.basic_block) > 3
                else None
            )
            if isinstance(maybe_compare, Instr) and maybe_compare.name == "COMPARE_OP":
                self._instrument_compare_op(node.basic_block)
            if (
                isinstance(maybe_string_func, Instr)
                and maybe_string_func.name == "LOAD_METHOD"
                and maybe_string_func.arg in self._STRING_FUNCTION_NAMES
            ):
                self._instrument_string_func(node.basic_block, maybe_string_func.arg)
            if (
                isinstance(maybe_string_func_with_arg, Instr)
                and maybe_string_func_with_arg.name == "LOAD_METHOD"
                and maybe_string_func_with_arg.arg in self._STRING_FUNCTION_NAMES
            ):
                self._instrument_string_func(
                    node.basic_block, maybe_string_func_with_arg.arg
                )

    def instrument_module(self, module_code: CodeType) -> CodeType:
        return self._instrument_code_recursive(module_code)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a factories for generating different kind of test cases."""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

import pynguin.analyses.seeding.initialpopulationseeding as initpopseeding
import pynguin.configuration as config
import pynguin.testcase.defaulttestcase as dtc
from pynguin.utils import randomness

if TYPE_CHECKING:
    import pynguin.testcase.testcase as tc
    import pynguin.testcase.testfactory as tf


# pylint:disable=too-few-public-methods
class TestCaseFactory:
    """Abstract class for test case factories."""

    def __init__(self, test_factory: tf.TestFactory):
        """Instantiates the factory.

        Args:
            test_factory: The used test factory
        """
        self._test_factory = test_factory

    @abstractmethod
    def get_test_case(self) -> tc.TestCase:
        """Retrieve a test case.

        Returns:
            A test case  # noqa: DAR202
        """


class RandomLengthTestCaseFactory(TestCaseFactory):
    """Create random test cases with random length."""

    def get_test_case(self) -> tc.TestCase:
        test_case = dtc.DefaultTestCase()
        attempts = 0
        size = randomness.next_int(
            1, config.configuration.search_algorithm.chromosome_length + 1
        )

        while (
            test_case.size() < size
            and attempts < config.configuration.test_creation.max_attempts
        ):
            self._test_factory.insert_random_statement(test_case, test_case.size())
            attempts += 1
        return test_case


class SeededTestCaseFactory(TestCaseFactory):
    """Factory for getting seeded test cases.

    With a certain probability a seeded testcase is returned instead of a randomly
    generated one. If a seeded testcase is returned, it is taken randomly from the
    pool of seeded testcases. If a randomly generated testcase is returned, the
    generation is delegated to the RandomLengthTestCaseFactory.
    """

    def __init__(self, delegate: TestCaseFactory, test_factory: tf.TestFactory):
        super().__init__(test_factory)
        self._delegate = delegate

    def get_test_case(self) -> tc.TestCase:
        if (
            config.configuration.seeding.initial_population_seeding
            and initpopseeding.initialpopulationseeding.has_tests
            and randomness.next_float()
            <= config.configuration.seeding.seeded_testcases_reuse_probability
        ):
            return initpopseeding.initialpopulationseeding.seeded_testcase
        return self._delegate.get_test_case()



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Utility methods for collections."""
from __future__ import annotations

from typing import Any


def dict_without_keys(dict_to_change: dict[Any, Any], keys: set[Any]) -> dict[Any, Any]:
    """
    Removes the given keys from the given dict.

    Args:
        dict_to_change: The dict where the keys should be removed.
        keys: The list of keys which should be removed.

    Returns:
        the dict without the specified keys.
    """
    return {k: v for k, v in dict_to_change.items() if k not in keys}



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a strategy implementation that utilises stub files."""
from __future__ import annotations

import ast
import inspect
import os
import sys
from pydoc import locate
from typing import Callable

from pynguin.typeinference.strategy import InferredSignature, TypeInferenceStrategy
from pynguin.typeinference.typehintsstrategy import TypeHintsInferenceStrategy


# pylint: disable=too-few-public-methods
class StubInferenceStrategy(TypeInferenceStrategy):
    """Provides a strategy that utilises stub files to infer variable types."""

    _cache: dict[str, ast.Module] = {}

    def __init__(self, pyi_dir: str | os.PathLike) -> None:
        self._pyi_dir = pyi_dir

    def infer_type_info(self, method: Callable) -> InferredSignature:
        assert self._pyi_dir

        module = sys.modules[method.__module__]
        name = module.__name__.replace(".", "/")
        pyi_src = name + ".pyi"
        pyi_ast = self._read_stub(pyi_src)
        parameter_types, return_type = self._get_parameter_annotations(method, pyi_ast)
        if parameter_types:
            return InferredSignature(
                signature=inspect.signature(method),
                parameters=parameter_types if parameter_types else {},
                return_type=return_type if return_type else None,
            )
        return TypeHintsInferenceStrategy().infer_type_info(method)

    def _read_stub(self, pyi_src: str) -> ast.Module | None:
        path = os.path.join(self._pyi_dir, pyi_src)
        if path in self._cache:
            return self._cache[path]

        try:
            with open(path, encoding="utf-8") as pyi_file:
                pyi_content = pyi_file.read()
                pyi_ast = ast.parse(pyi_content)
                self._cache[path] = pyi_ast
        except OSError:
            return None
        return pyi_ast

    @staticmethod
    def _get_parameter_annotations(
        method: Callable, pyi_ast: ast.Module | None
    ) -> tuple[dict[str, type | None], type | None]:
        if not pyi_ast:
            return {}, None

        param_types_matches: dict[str, type | None] = {}
        return_type: type | None = None
        if inspect.isfunction(method):
            function_listener = _FunctionListener(pyi_ast, method.__name__)
            function_listener.visit(pyi_ast)
            param_types_matches = function_listener.param_type_matches
            return_type = function_listener.return_type
        elif inspect.isclass(method):
            class_listener = _ClassListener(pyi_ast, method.__name__, "__init__")
            class_listener.visit(pyi_ast)
            param_types_matches = class_listener.param_type_matches
        return param_types_matches, return_type


class _FunctionListener(ast.NodeVisitor):
    def __init__(self, pyi_ast: ast.Module | None, name: str) -> None:
        self._pyi_ast = pyi_ast
        self._name = name
        self._param_type_matches: dict[str, type | None] = {}
        self._return_type: type | None = None

    @property
    def param_type_matches(self) -> dict[str, type | None]:
        """Provides the matched parameter types.

        Returns:
            The matched parameter types
        """
        if "self" in self._param_type_matches:
            del self._param_type_matches["self"]
        return self._param_type_matches

    @property
    def return_type(self) -> type | None:
        """Provides the return type.

        Returns:
            Teh inferred return type
        """
        return self._return_type

    # pylint: disable=invalid-name, missing-function-docstring
    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        if self._name == node.name:
            self._param_type_matches.clear()
            for name, field in ast.iter_fields(node):
                if name == "args":
                    for arg in field.args:
                        if arg.annotation and hasattr(arg.annotation, "id"):
                            self._param_type_matches[arg.arg] = self._locate(
                                arg.annotation.id
                            )
                        elif arg.annotation and hasattr(arg.annotation, "attr"):
                            self._param_type_matches[arg.arg] = self._locate(
                                f"{arg.annotation.value.id}.{arg.annotation.attr}"
                            )
                        else:
                            self._param_type_matches[arg.arg] = None
            if node.returns and hasattr(node.returns, "id"):
                self._return_type = self._locate(node.returns.id)  # type: ignore

    def _locate(self, identifier: str | None) -> type | None:
        assert self._pyi_ast
        if not identifier:
            return None
        result = locate(identifier)
        if result:
            return result  # type: ignore
        visitor = _ImportVisitor(identifier)
        visitor.generic_visit(self._pyi_ast)
        return locate(visitor.full_qualified_name)  # type: ignore


class _ImportVisitor(ast.NodeVisitor):
    def __init__(self, name: str) -> None:
        self._name = name
        self._full_qualified_name: str | None = None

    @property
    def full_qualified_name(self) -> str | None:
        """Provides the inferred fully-qualified name.

        Returns:
            The fully-qualified name
        """
        return self._full_qualified_name

    # pylint: disable=invalid-name, missing-function-docstring
    def visit_ImportFrom(self, statement: ast.ImportFrom) -> None:
        if statement.names:
            for alias in statement.names:
                if alias.name == self._name:
                    self._full_qualified_name = f"{statement.module}.{alias.name}"


class _ClassListener(ast.NodeVisitor):
    def __init__(
        self, pyi_ast: ast.Module | None, class_name: str, function_name: str
    ) -> None:
        self._pyi_ast = pyi_ast
        self._class_name = class_name
        self._function_name = function_name
        self._param_type_matches: dict[str, type | None] = {}

    @property
    def param_type_matches(self) -> dict[str, type | None]:
        """Provides the matched parameter types.

        Returns:
            The matched parameter types
        """
        return self._param_type_matches

    # pylint: disable=invalid-name, missing-function-docstring
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if self._class_name == node.name:
            listener = _FunctionListener(self._pyi_ast, self._function_name)
            listener.generic_visit(node)
            self._param_type_matches = listener.param_type_matches



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
"""Provide tournament selection."""
from __future__ import annotations

from typing import TypeVar

import pynguin.configuration as config
import pynguin.ga.chromosome as chrom
from pynguin.ga.operators.selection.selection import SelectionFunction
from pynguin.utils import randomness

T = TypeVar("T", bound=chrom.Chromosome)  # pylint: disable=invalid-name


class TournamentSelection(SelectionFunction[T]):
    """Tournament selection."""

    def get_index(self, population: list[T]) -> int:
        new_num = randomness.next_int(lower_bound=0, upper_bound=len(population))
        winner = new_num

        tournament_round = 0

        while (
            tournament_round < config.configuration.search_algorithm.tournament_size - 1
        ):
            new_num = randomness.next_int(lower_bound=0, upper_bound=len(population))
            selected = population[new_num]

            if self._maximize:
                if selected.get_fitness() > population[winner].get_fitness():
                    winner = new_num
            else:
                if selected.get_fitness() < population[winner].get_fitness():
                    winner = new_num

            tournament_round += 1

        return winner



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides classes for handling fitness functions for branch coverage."""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from ordered_set import OrderedSet

import pynguin.coverage.controlflowdistance as cfd
import pynguin.ga.computations as ff

if TYPE_CHECKING:
    import pynguin.ga.testcasechromosome as tcc
    from pynguin.testcase.execution import (
        ExecutionResult,
        ExecutionTracer,
        KnownData,
        TestCaseExecutor,
    )


class AbstractBranchCoverageGoal:
    """Abstract base class for branch coverage goals."""

    def __init__(
        self,
        code_object_id: int,
        is_branchless_code_object: bool = False,
        is_branch: bool = False,
    ):
        assert (
            is_branchless_code_object ^ is_branch
        ), "Must be either branch-less code object or branch."
        self._is_branchless_code_object = is_branchless_code_object
        self._is_branch = is_branch
        self._code_object_id = code_object_id

    @property
    def code_object_id(self) -> int:
        """Provides the code object id where the target resides.

        Returns:
            The id of the targeted code object.
        """
        return self._code_object_id

    @abstractmethod
    def get_distance(
        self, result: ExecutionResult, tracer: ExecutionTracer
    ) -> cfd.ControlFlowDistance:
        """Computes the control-flow distance of an execution result.

        Args:
            result: The execution result
            tracer: The execution tracer

        Returns:
            The control-flow distance
        """

    @abstractmethod
    def is_covered(self, result: ExecutionResult) -> bool:
        """Determine if this branch coverage goal was covered.
        Faster to compute than get_distance but
        doesn't give any guidance as it returns only a bool.

        Args:
            result: The execution result to check.

        Returns:
            True, if this goal is covered in the execution result
        """

    @property
    def is_branchless_code_object(self) -> bool:
        """Does this target a branch-less code object?

        Returns:
            True, if it targets a branch-less code object.
        """
        return self._is_branchless_code_object

    @property
    def is_branch(self) -> bool:
        """Does this target a certain execution of a predicate?

        Returns:
            True, if it targets an execution of a predicate.
        """
        return self._is_branch


class BranchlessCodeObjectGoal(AbstractBranchCoverageGoal):
    """Entry into a code object without branches."""

    def __init__(self, code_object_id: int):
        super().__init__(code_object_id=code_object_id, is_branchless_code_object=True)

    def get_distance(
        self, result: ExecutionResult, tracer: ExecutionTracer
    ) -> cfd.ControlFlowDistance:
        return cfd.get_root_control_flow_distance(result, self._code_object_id, tracer)

    def is_covered(self, result: ExecutionResult) -> bool:
        return self._code_object_id in result.execution_trace.executed_code_objects

    def __str__(self) -> str:
        return f"Branch-less Code-Object {self._code_object_id}"

    def __repr__(self) -> str:
        return f"BranchlessCodeObjectGoal({self._code_object_id})"

    def __hash__(self) -> int:
        return 31 + self._code_object_id

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, BranchlessCodeObjectGoal):
            return False
        return self._code_object_id == other._code_object_id


class BranchGoal(AbstractBranchCoverageGoal):
    """The true/false evaluation of a jump condition."""

    def __init__(self, code_object_id: int, predicate_id: int, value: bool):
        super().__init__(code_object_id=code_object_id, is_branch=True)
        self._predicate_id = predicate_id
        self._value = value

    def get_distance(
        self, result: ExecutionResult, tracer: ExecutionTracer
    ) -> cfd.ControlFlowDistance:
        return cfd.get_non_root_control_flow_distance(
            result, self._predicate_id, self._value, tracer
        )

    def is_covered(self, result: ExecutionResult) -> bool:
        trace = result.execution_trace
        distances = trace.true_distances if self._value else trace.false_distances
        return (
            self._predicate_id in trace.executed_predicates
            and distances[self._predicate_id] == 0.0
        )

    @property
    def predicate_id(self) -> int:
        """Provides the predicate id of the targeted predicate.

        Returns:
            The id of the targeted predicate.
        """
        return self._predicate_id

    @property
    def value(self) -> bool:
        """Provides whether we target the True or False branch of the predicate.

        Returns:
            The targeted branch value.
        """
        return self._value

    def __str__(self) -> str:
        return f"{self._value} branch of predicate {self._predicate_id}"

    def __repr__(self) -> str:
        return f"BranchGoal(predicate_id={self._predicate_id}, " f"value={self._value})"

    def __hash__(self) -> int:
        prime = 31
        result = 1
        result = prime * result + self._predicate_id
        result = prime * result + int(self._value)
        return result

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, BranchGoal):
            return False
        return self.predicate_id == other.predicate_id and self._value == other.value


class BranchGoalPool:
    """Convenience class that creates and provides all branch coverage related goals."""

    def __init__(self, known_data: KnownData):
        self._branchless_code_object_goals = self._compute_branchless_code_object_goals(
            known_data
        )
        self._predicate_to_branch_goals = self._compute_branch_goals(known_data)

    @property
    def branchless_code_object_goals(self) -> list[BranchlessCodeObjectGoal]:
        """Provide the goals for branch-less code objects.

        Returns:
            The goals for branch-less code objects.
        """
        return self._branchless_code_object_goals

    @property
    def branch_goals(self) -> list[BranchGoal]:
        """Provide the goals for branches.

        Returns:
            The goals for branches.
        """
        return [
            goal for goals in self._predicate_to_branch_goals.values() for goal in goals
        ]

    @property
    def branch_coverage_goals(self) -> OrderedSet[AbstractBranchCoverageGoal]:
        """Provide all goals related to branch coverage.

        Returns:
            All goals related to branch coverage.
        """
        goals: OrderedSet[AbstractBranchCoverageGoal] = OrderedSet(self.branch_goals)
        goals.update(self.branchless_code_object_goals)
        return goals

    @staticmethod
    def _compute_branchless_code_object_goals(
        known_data: KnownData,
    ) -> list[BranchlessCodeObjectGoal]:
        return [
            BranchlessCodeObjectGoal(code_object_id)
            for code_object_id in known_data.branch_less_code_objects
        ]

    @staticmethod
    def _compute_branch_goals(known_data: KnownData) -> dict[int, list[BranchGoal]]:
        goal_map: dict[int, list[BranchGoal]] = {}
        for predicate_id, meta in known_data.existing_predicates.items():
            entry: list[BranchGoal] = []
            goal_map[predicate_id] = entry
            entry.append(BranchGoal(meta.code_object_id, predicate_id, True))
            entry.append(BranchGoal(meta.code_object_id, predicate_id, False))
        return goal_map


class BranchCoverageTestFitness(ff.TestCaseFitnessFunction):
    """A branch coverage fitness implementation for test cases."""

    def __init__(self, executor: TestCaseExecutor, goal: AbstractBranchCoverageGoal):
        super().__init__(executor, goal.code_object_id)
        self._goal = goal

    def compute_fitness(self, individual: tcc.TestCaseChromosome) -> float:
        result = self._run_test_case_chromosome(individual)

        distance = self._goal.get_distance(result, self._executor.tracer)
        return distance.get_resulting_branch_fitness()

    def compute_is_covered(self, individual: tcc.TestCaseChromosome) -> bool:
        result = self._run_test_case_chromosome(individual)
        return self._goal.is_covered(result)

    def is_maximisation_function(self) -> bool:
        return False

    def __str__(self) -> str:
        return f"BranchCoverageTestFitness for {self._goal}"

    def __repr__(self) -> str:
        return (
            f"BranchCoverageTestFitness(executor={self._executor}, "
            f"goal={self._goal})"
        )

    @property
    def goal(self) -> AbstractBranchCoverageGoal:
        """Provides the branch-coverage goal of this fitness function.

        Returns:
            The attached branch-coverage goal
        """
        return self._goal


def create_branch_coverage_fitness_functions(
    executor: TestCaseExecutor, branch_goal_pool: BranchGoalPool
) -> OrderedSet[BranchCoverageTestFitness]:
    """Create fitness functions for each branch coverage goal.

    Args:
        executor: The test case executor for the fitness functions to use.
        branch_goal_pool: The pool that holds all branch goals.

    Returns:
        All branch coverage related fitness functions.
    """
    return OrderedSet(
        [
            BranchCoverageTestFitness(executor, goal)
            for goal in branch_goal_pool.branch_coverage_goals
        ]
    )



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
"""Provides classes for the internal representation of types."""
from __future__ import annotations

import functools
import math
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, Iterable, Iterator, NamedTuple

from pynguin.utils import randomness

if TYPE_CHECKING:
    from pynguin.analyses.module.inheritance import ClassInformation, InheritanceGraph


# pylint: disable=too-few-public-methods
class SignatureType(metaclass=ABCMeta):
    """An abstract base class for types in signatures."""


# pylint: disable=too-few-public-methods
class _UnknownType(SignatureType):
    """Represents an unknown type, i.e., a type without an annotation."""


# pylint: disable=too-few-public-methods
class _AnyType(SignatureType):
    """Represents any type, i.e., a type annotated with Any."""


class ConcreteType(SignatureType):
    """Represents a concrete type."""

    def __init__(self, class_information: ClassInformation) -> None:
        self._class_information = class_information

    @property
    def class_information(self) -> ClassInformation:
        """Provides the class information of this type.

        Returns:
            The class information
        """
        return self._class_information

    @property
    def type_name(self) -> str:
        """Provides the name of the class.

        Returns:
            The name of the class
        """
        return self._class_information.name

    @property
    def type_object(self) -> type:
        """Provides the type of the class.

        Returns:
            The type of the class
        """
        return self._class_information.class_object

    def __str__(self) -> str:
        return f"ConcreteType({self._class_information})"

    def __repr__(self) -> str:
        return f"ConcreteType({self._class_information!r})"

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, ConcreteType):
            return False
        return self._class_information == other.class_information

    def __hash__(self) -> int:
        return hash(self._class_information)


class SignatureElement(metaclass=ABCMeta):
    """An abstract base class for types in signatures."""

    @functools.total_ordering
    class Element(NamedTuple):
        """Represents an element."""

        signature_type: SignatureType
        confidence: float

        def __eq__(self, other: Any) -> bool:
            if self is other:
                return True
            if not isinstance(other, SignatureElement.Element):
                return False
            return self.signature_type == other.signature_type and math.isclose(
                self.confidence, other.confidence, rel_tol=1e-6, abs_tol=1e-12
            )

        def __lt__(self, other: Any) -> bool:
            if not isinstance(other, SignatureElement.Element):
                raise TypeError(  # pylint: disable=raising-format-tuple
                    "'<' not supported between instances of "
                    "'SignatureElement._Element' and '%s'",
                    type(other),
                )
            return self.confidence < other.confidence

        def __repr__(self) -> str:
            return f"Element({self.signature_type!r}, {self.confidence})"

        def __str__(self) -> str:
            return f"Element({self.signature_type}, {self.confidence})"

    _related_type_confidence: ClassVar[float] = 0.5  # pylint: disable=invalid-name

    def __init__(self) -> None:
        self._unknown_element = self.Element(
            signature_type=unknown_type, confidence=0.0
        )
        self._elements: set[SignatureElement.Element] = {self._unknown_element}
        self._inheritance_graph: InheritanceGraph | None = None

    def add_element(self, signature: SignatureType, confidence: float) -> None:
        """Adds an element to the set of possible signature types.

        The confidence value must be from [0;1], otherwise a `ValueError` is raised.
        An element must not be added twice.

        Args:
            signature: The element to add
            confidence: Its confidence

        Raises:
            AssertionError: If the element to be added already exists
            ValueError: If the confidence level is not from [0;1]
        """
        if confidence > 1 or confidence < 0:
            raise ValueError("Confidence must be in [0;1].")
        if self._contains_signature(signature):
            raise AssertionError(
                "It is illegal to add an element of the same type twice.  If you want "
                "to update the element's confidence, use `replace_element`."
            )
        if len(self._elements) == 1 and self._unknown_element in self._elements:
            self._elements.clear()
        element = self._element_factory(signature, confidence)
        self._elements.add(element)

    def _contains_signature(self, signature: SignatureType) -> bool:
        for element in self._elements:
            if element.signature_type == signature:
                return True
        return False

    def _element_factory(
        self, signature: SignatureType, confidence: float
    ) -> SignatureElement.Element:
        return self.Element(signature_type=signature, confidence=confidence)

    def replace_element(self, signature: SignatureType, confidence: float) -> None:
        """Replace the elements confidence.

        If the element is not yet present it will be added.
        The confidence value must be from [0;1], otherwise a `ValueError` is raised.

        Args:
            signature: The element to consider
            confidence: The new confidence value

        Raises:
            ValueError: If the confidence level is not from [0;1]
        """
        if confidence > 1 or confidence < 0:
            raise ValueError("Confidence must be in [0;1].")
        if not self._contains_signature(signature):
            self.add_element(signature, confidence)
        else:
            found: SignatureElement.Element | None = self.get_element(signature)
            assert found is not None
            self._elements.remove(found)
            self._elements.add(self._element_factory(signature, confidence))

    def get_element(self, signature: SignatureType) -> SignatureElement.Element | None:
        """Provides the element with a given signature type.

        Returns None if not found

        Args:
            signature: The signature type to search for

        Returns:
            The element with this signature type, or None if not found
        """
        for element in self._elements:
            if element.signature_type == signature:
                return element
        return None

    def provide_random_type(self, respect_confidence: bool = True) -> SignatureType:
        """Provides a random type from the possible types.

        If the `respect_confidence` parameter is set, it will sample based on the
        confidence level, otherwise it will randomly choose from all types.

        Args:
            respect_confidence: Whether or not the confidence level shall be respected

        Returns:
            A random signature type
        """
        assert len(self._elements) > 0
        if not respect_confidence:
            return randomness.choice(tuple(self._elements)).signature_type
        # use the fact that zip behaves almost like its own inverse to unzip the set
        # of pairs to two sequences.  Seems like magic but is actually a nice thing.
        signatures, confidences = tuple(
            zip(
                *[
                    (element.signature_type, element.confidence)
                    for element in self._elements
                ]
            )
        )
        return randomness.choices(signatures, weights=confidences)[0]

    @property
    def elements(self) -> set[SignatureElement.Element]:
        """Provides all types that are possible.

        Returns:
            All types that are possible
        """
        return self._elements

    @property
    def element_types(self) -> Iterator[SignatureType]:
        """Provides an iterator over the element types.

        Returns:
            An iterator over the element types
        """
        for sig_type, _ in self._elements:
            yield sig_type

    @abstractmethod
    def include_inheritance(self, inheritance_graph: InheritanceGraph) -> None:
        """Includes the information from an inheritance graph.

        Args:
            inheritance_graph: The inheritance graph to use
        """

    def _update_inherited_types(
        self,
        element_types: Iterable[SignatureType],
        start_type: ClassInformation,
        inherited_types: Iterable[ClassInformation],
    ) -> None:
        assert self._inheritance_graph is not None
        for inherited_type in inherited_types:
            concrete_inherited_type = ConcreteType(inherited_type)
            distance = self._inheritance_graph.get_distance(start_type, inherited_type)
            confidence = self._related_type_confidence ** abs(distance)
            if concrete_inherited_type not in element_types:
                self.add_element(concrete_inherited_type, confidence)
            else:
                existing = self.get_element(concrete_inherited_type)
                assert existing is not None
                self.replace_element(
                    existing.signature_type, max(existing.confidence, confidence)
                )


class Parameter(SignatureElement):
    """Represents a parameter of a method."""

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    @property
    def name(self) -> str:
        """Provides the parameter name.

        Returns:
            The parameter name
        """
        return self._name

    def include_inheritance(self, inheritance_graph: InheritanceGraph) -> None:
        self._inheritance_graph = inheritance_graph
        element_types = list(self.element_types)
        for element_type in element_types:
            if isinstance(element_type, ConcreteType):
                sub_types = inheritance_graph.get_sub_types(
                    element_type.class_information
                )
                self._update_inherited_types(
                    element_types, element_type.class_information, sub_types
                )


class ReturnType(SignatureElement):
    """Represents the return type of a method."""

    def include_inheritance(self, inheritance_graph: InheritanceGraph) -> None:
        self._inheritance_graph = inheritance_graph
        element_types = list(self.element_types)
        for element_type in element_types:
            if isinstance(element_type, ConcreteType):
                super_types = inheritance_graph.get_super_types(
                    element_type.class_information
                )
                self._update_inherited_types(
                    element_types, element_type.class_information, super_types
                )


unknown_type = _UnknownType()
any_type = _AnyType()



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides rich's Console handler class."""
from rich.console import Console

console = Console(tab_size=4)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a random test generator, that creates random test suites."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pynguin.generation.algorithms.testgenerationstrategy import TestGenerationStrategy

if TYPE_CHECKING:
    import pynguin.ga.testsuitechromosome as tsc


# pylint: disable=too-few-public-methods
class RandomTestSuiteSearchStrategy(TestGenerationStrategy):
    """Create random test suites."""

    _logger = logging.getLogger(__name__)

    def generate_tests(
        self,
    ) -> tsc.TestSuiteChromosome:
        self.before_search_start()
        solution = self._chromosome_factory.get_chromosome()
        self.before_first_search_iteration(solution)
        while self.resources_left() and solution.get_fitness() != 0.0:
            candidate = self._chromosome_factory.get_chromosome()
            if candidate.get_fitness() < solution.get_fitness():
                solution = candidate
            self.after_search_iteration(solution)
        self.after_search_finish()
        return solution


class RandomTestCaseSearchStrategy(TestGenerationStrategy):
    """Creates random test suites based on test-case chromosomes."""

    _logger = logging.getLogger(__name__)

    def generate_tests(self) -> tsc.TestSuiteChromosome:
        self.before_search_start()
        solution = self._chromosome_factory.get_chromosome()
        self._archive.update([solution])
        test_suite = self.create_test_suite(self._archive.solutions)
        self.before_first_search_iteration(test_suite)
        while self.resources_left() and test_suite.get_fitness() != 0.0:
            candidate = self._chromosome_factory.get_chromosome()
            self._archive.update([candidate])
            test_suite = self.create_test_suite(self._archive.solutions)
            self.after_search_iteration(test_suite)
        self.after_search_finish()
        return self.create_test_suite(self._archive.solutions)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2020 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides some mutation related utilities."""
from __future__ import annotations

from typing import Callable, TypeVar

from pynguin.utils import randomness

T = TypeVar("T")  # pylint:disable=invalid-name


def alpha_exponent_insertion(
    elements: list[T],
    value_supplier: Callable[[], T | None],
    alpha: float = 0.5,
    exponent: int = 0,
) -> bool:
    """Repeatedly inserts a new element generated by value_supplier into the given
    elements at a random position as long as next_float() < alpha^exponent holds.
    Exponent is increased after each insertion, thus lowering the chance for another
    insertion.

    Args:
        elements: the elements into which new elements should be inserted.
        value_supplier: supplies the elements that are inserted.
        alpha: the used alpha value.
        exponent: start value of the exponent.

    Returns:
        True, iff at least one element was inserted.
    """
    assert 0 < alpha < 1
    pos = 0
    changed = False
    while randomness.next_float() <= pow(alpha, exponent):
        # Randomize the position for each insertion.
        if len(elements) > 0:
            pos = randomness.next_int(0, len(elements) + 1)

        exponent += 1
        value = value_supplier()
        if value is None:
            # Supplier is exhausted
            return changed

        elements.insert(
            pos,
            value,
        )
        changed = True

    return changed



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides abstract factories for output variables"""
from __future__ import annotations

import time
import typing
from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar, overload

import pynguin.configuration as config
import pynguin.utils.statistics.statisticsbackend as sb

if typing.TYPE_CHECKING:
    import pynguin.ga.chromosome as chrom
    import pynguin.utils.statistics.statistics as stat
    from pynguin.utils.statistics.runtimevariable import RuntimeVariable

T = TypeVar("T", int, float)  # pylint: disable=invalid-name


class ChromosomeOutputVariableFactory(Generic[T], metaclass=ABCMeta):
    """Factory to create an output variable when given a test suite chromosome"""

    def __init__(self, variable: RuntimeVariable) -> None:
        self._variable = variable

    @abstractmethod
    def get_data(self, individual: chrom.Chromosome) -> T:
        """Returns the data value from the individual.

        Args:
            individual: The individual to query

        Returns:
            The current value of the variable in the individual  # noqa: DAR202
        """

    def get_variable(self, individual: chrom.Chromosome) -> sb.OutputVariable[T]:
        """Provides the output variable

        Args:
            individual: The individual

        Returns:
            The output variable for the individual
        """
        return sb.OutputVariable(
            name=self._variable.name, value=self.get_data(individual)
        )


class SequenceOutputVariableFactory(Generic[T], metaclass=ABCMeta):
    """Creates an output variable that represents a sequence of values"""

    def __init__(self, variable: stat.RuntimeVariable) -> None:
        self._variable = variable
        self._time_stamps: list[int] = []
        self._values: list[T] = []
        self._start_time: int = 0

    def set_start_time(self, start_time: int) -> None:
        """Sets the start time.

        Args:
            start_time: the start time
        """
        self._start_time = start_time

    @abstractmethod
    def get_value(self, individual: chrom.Chromosome) -> T:
        """Returns the current value of the variable for the selected individual

        Args:
            individual: The individual to query

        Returns:
            The current value of the variable in the individual  # noqa: DAR202
        """

    def update(self, individual: chrom.Chromosome) -> None:
        """Updates the values for an individual

        Args:
            individual: The individual
        """
        self._time_stamps.append(time.time_ns() - self._start_time)
        self._values.append(self.get_value(individual))

    @overload
    def update_value(self, value: int) -> None:
        ...

    @overload
    def update_value(self, value: float) -> None:
        ...

    def update_value(self, value) -> None:
        """Updates the value directly.

        Args:
            value: The value
        """
        self._time_stamps.append(time.time_ns() - self._start_time)
        self._values.append(value)

    def get_variable_names_indices(self) -> list[tuple[int, str]]:
        """Provides a list of variable names

        Returns:
            A list of pairs consisting of variable names and their index.
        """
        return [
            (i + 1, f"{self._variable.name}_T{i + 1}")
            for i in range(self._calculate_number_of_intervals())
        ]

    def get_output_variables(self) -> list[sb.OutputVariable[T]]:
        """Provides the output variables

        Returns:
            A list of output variables
        """
        return [
            sb.OutputVariable(
                name=variable_name, value=self._get_time_line_value(variable_index)
            )
            for variable_index, variable_name in self.get_variable_names_indices()
        ]

    def _get_time_line_value(self, index: int) -> T:
        if not self._time_stamps:
            # No data, if this is even possible.
            return 0
        interval = config.configuration.statistics_output.timeline_interval
        preferred_time = interval * index
        # pylint:disable=consider-using-enumerate
        for i in range(len(self._time_stamps)):
            # find the first stamp that is following the time we would like to get
            # the value for
            stamp = self._time_stamps[i]
            if stamp < preferred_time:
                continue

            if i == 0:
                # it is the first element, just use it as value
                return self._values[i]

            if not config.configuration.statistics_output.timeline_interpolation:
                # if we do not want to interpolate, return last observed value
                return self._values[i - 1]

            # interpolate the value, since we do not have the value for the exact
            # time we want
            time_delta = self._time_stamps[i] - self._time_stamps[i - 1]
            if time_delta > 0:
                value_delta = float(self._values[i]) - float(self._values[i - 1])
                ratio = value_delta / time_delta
                diff = preferred_time - self._time_stamps[i - 1]
                value = float(self._values[i - 1]) + (diff * ratio)
                return value  # type: ignore

        # no time stamp was higher, just use the last value seen
        return self._values[-1]

    @staticmethod
    def _calculate_number_of_intervals() -> int:
        interval = config.configuration.statistics_output.timeline_interval
        total_time = config.configuration.stopping.budget * 1_000_000_000
        number_of_intervals = total_time // interval
        return int(number_of_intervals)


class DirectSequenceOutputVariableFactory(SequenceOutputVariableFactory, Generic[T]):
    """Sequence output variable whose value can be set directly, instead of
    retrieving it from an individual"""

    def __init__(self, variable: RuntimeVariable, start_value: T) -> None:
        super().__init__(variable)
        self._value = start_value  # type: ignore

    def get_value(self, individual) -> T:
        return self._value

    def set_value(self, value: T) -> None:
        """Sets the value directly.

        Args:
            value: the value to be set
        """
        self._value = value

    @staticmethod
    def get_float(
        variable: RuntimeVariable,
    ) -> DirectSequenceOutputVariableFactory:
        """Creates a factory for a float variable.

        Args:
            variable: the runtime variable

        Returns:
            A factory for that variable
        """
        return DirectSequenceOutputVariableFactory(variable, 0.0)

    @staticmethod
    def get_integer(
        variable: RuntimeVariable,
    ) -> DirectSequenceOutputVariableFactory:
        """Creates a factory for an integer variable.

        Args:
            variable: the runtime variable

        Returns:
            A factory for that variable
        """
        return DirectSequenceOutputVariableFactory(variable, 0)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Factory for chromosome used by the genetic algorithm."""
from abc import abstractmethod
from typing import Generic, TypeVar

import pynguin.ga.chromosome as chrom

T = TypeVar("T", bound=chrom.Chromosome)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class ChromosomeFactory(Generic[T]):
    """A factory that provides new chromosomes."""

    @abstractmethod
    def get_chromosome(self) -> T:
        """Create a new chromosome.

        Returns:
            A new chromosome  # noqa: DAR202
        """



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Pynguin is an automated unit test generation framework for Python."""
import pynguin.configuration as config
import pynguin.generator as gen

set_configuration = gen.set_configuration
run_pynguin = gen.run_pynguin
Configuration = config.Configuration
Algorithm = config.Algorithm
ExportStrategy = config.ExportStrategy
StatisticsBackend = config.StatisticsBackend
StoppingCondition = config.StoppingCondition
TypeInferenceStrategy = config.TypeInferenceStrategy

__version__ = "0.16.1"
__all__ = [
    "set_configuration",
    "run_pynguin",
    "Configuration",
    "__version__",
    "Algorithm",
    "ExportStrategy",
    "StatisticsBackend",
    "StoppingCondition",
    "TypeInferenceStrategy",
]



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides a configuration interface for the test generator."""
import dataclasses
import enum
from typing import List, Optional

from pynguin.utils.statistics.runtimevariable import RuntimeVariable


class ExportStrategy(str, enum.Enum):
    """Contains all available export strategies.

    These strategies allow to export the generated test cases in different styles,
    such as the style of the `PyTest` framework.  Setting the value to `NONE` will
    prevent exporting of the generated test cases (only reasonable for
    benchmarking, though).
    """

    PY_TEST = "PY_TEST"
    """Export tests in the style of the PyTest framework."""

    NONE = "NONE"
    """Do not export test cases at all."""


class Algorithm(str, enum.Enum):
    """Different algorithms supported by Pynguin."""

    DYNAMOSA = "DYNAMOSA"
    """The dynamic many-objective sorting algorithm (cf. Panichella et al. Automated
    test case generation as a many-objective optimisation problem with dynamic selection
    of the targets.  TSE vol. 44 issue 2)."""

    MIO = "MIO"
    """The MIO test suite generation algorithm (cf. Andrea Arcuri. Many Independent
    Objective (MIO) Algorithm for Test Suite Generation.  Proc. SBSE 2017)."""

    MOSA = "MOSA"
    """The many-objective sorting algorithm (cf. Panichella et al. Reformulating Branch
    Coverage as a Many-Objective Optimization Problem.  Proc. ICST 2015)."""

    RANDOM = "RANDOM"
    """A feedback-direct random test generation approach similar to the algorithm
    proposed by Randoop (cf. Pacheco et al. Feedback-directed random test generation.
    Proc. ICSE 2007)."""

    RANDOM_TEST_SUITE_SEARCH = "RANDOM_TEST_SUITE_SEARCH"
    """Performs random search on test suites."""

    RANDOM_TEST_CASE_SEARCH = "RANDOM_TEST_CASE_SEARCH"
    """Performs random search on test cases."""

    WHOLE_SUITE = "WHOLE_SUITE"
    """A whole-suite test generation approach similar to the one proposed by EvoSuite
    (cf. Fraser and Arcuri. EvoSuite: Automatic Test Suite Generation for
    Object-Oriented Software. Proc. ESEC/FSE 2011).

    This algorithm can be modified to use an archive (cf. Rojas, José Miguel, et al.
    "A detailed investigation of the effectiveness of whole test suite generation."
    Empirical Software Engineering 22.2 (2017): 852-893.), by using the
    following options: --use-archive True, --seed_from_archive True and
    --filter_covered_targets_from_test_cluster True.
    """


class AssertionGenerator(str, enum.Enum):
    """Different approaches for assertion generation supported by Pynguin."""

    MUTATION_ANALYSIS = "MUTATION_ANALYSIS"
    """Use the mutation analysis approach for assertion generation."""

    SIMPLE = "SIMPLE"
    """Use the simple approach for primitive and none assertion generation."""

    NONE = "NONE"
    """Do not create any assertions."""


class MutationStrategy(str, enum.Enum):
    """Different strategies for creating mutants when using the MUTATION_ANALYSIS
    approach for assertion generation."""

    FIRST_ORDER_MUTANTS = "FIRST_ORDER_MUTANTS"
    """Generate first order mutants."""

    FIRST_TO_LAST = "FIRST_TO_LAST"
    """Higher order mutation strategy FirstToLast.
    (cf. Mateo et al. Validating Second-Order Mutation at System Level. Article.
    IEEE Transactions on SE 39.4 2013)"""

    BETWEEN_OPERATORS = "BETWEEN_OPERATORS"
    """Higher order mutation strategy BetweenOperators.
    (cf. Mateo et al. Validating Second-Order Mutation at System Level. Article.
    IEEE Transactions on SE 39.4 2013)"""

    RANDOM = "RANDOM"
    """Higher order mutation strategy Random.
    (cf. Mateo et al. Validating Second-Order Mutation at System Level. Article.
    IEEE Transactions on SE 39.4 2013)"""

    EACH_CHOICE = "EACH_CHOICE"
    """Higher order mutation strategy EachChoice.
    (cf. Mateo et al. Validating Second-Order Mutation at System Level. Article.
    IEEE Transactions on SE 39.4 2013)"""


class StoppingCondition(str, enum.Enum):
    """The different stopping conditions for the algorithms."""

    MAX_TIME = "MAX_TIME"
    """Stop after a maximum time limit has been reached."""

    MAX_ITERATIONS = "MAX_ITERATIONS"
    """Stop after a maximum number of algorithm iterations."""

    MAX_TEST_EXECUTIONS = "MAX_TEST_EXECUTIONS"
    """Stop as soon as a maximum number of tests were executed."""

    MAX_STATEMENT_EXECUTIONS = "MAX_STATEMENT_EXECUTIONS"
    """Stop as soon as a maximum number of statements were executed."""


class TypeInferenceStrategy(str, enum.Enum):
    """The different available type-inference strategies."""

    NONE = "NONE"
    """Ignore any type information given in the module under test."""

    STUB_FILES = "STUB_FILES"
    """Use type information from stub files."""

    TYPE_HINTS = "TYPE_HINTS"
    """Use type information from type hints in the module under test."""


class StatisticsBackend(str, enum.Enum):
    """The different available statistics backends to write statistics"""

    NONE = "NONE"
    """Do not write any statistics."""

    CONSOLE = "CONSOLE"
    """Write statistics to the standard out."""

    CSV = "CSV"
    """Write statistics to a CSV file."""


class Selection(str, enum.Enum):
    """Different selection algorithms to select from."""

    RANK_SELECTION = "RANK_SELECTION"
    """Rank selection."""

    TOURNAMENT_SELECTION = "TOURNAMENT_SELECTION"
    """Tournament selection.  Use `tournament_size` to set size."""


# pylint:disable=too-many-instance-attributes
@dataclasses.dataclass
class StatisticsOutputConfiguration:
    """Configuration related to output."""

    report_dir: str = "pynguin-report"
    """Directory in which to put HTML and CSV reports"""

    statistics_backend: StatisticsBackend = StatisticsBackend.CSV
    """Which backend to use to collect data"""

    timeline_interval: int = 1 * 1_000_000_000
    """Time interval in nano-seconds for timeline statistics, i.e., we select a data
    point after each interval.  This can be interpolated, if there is no exact
    value stored at the time-step of the interval, see `timeline_interpolation`.
    The default value is every 1.00s."""

    timeline_interpolation: bool = True
    """Interpolate timeline values"""

    output_variables: List[RuntimeVariable] = dataclasses.field(
        default_factory=lambda: [
            RuntimeVariable.TargetModule,
            RuntimeVariable.Coverage,
        ]
    )
    """List of variables to output to the statistics backend."""

    configuration_id: str = ""
    """Label that identifies the used configuration of Pynguin.  This is only done
    when running experiments."""

    project_name: str = ""
    """Label that identifies the project name of Pynguin.  This is useful when
    running experiments."""

    create_coverage_report: bool = False
    """Create a coverage report for the tested module.
    This can be helpful to find hard to cover parts because Pynguin measures coverage
    on bytecode level which might yield different results when compared with other
    tools, e.g., Coverage.py."""


@dataclasses.dataclass
class TestCaseOutputConfiguration:
    """Configuration related to test case output."""

    output_path: str
    """Path to an output folder for the generated test cases."""

    export_strategy: ExportStrategy = ExportStrategy.PY_TEST
    """The export strategy determines for which test-runner system the
    generated tests should fit."""

    max_length_test_case: int = 2500
    """The maximum number of statement in as test case (normal + assertion
    statements)"""

    assertion_generation: AssertionGenerator = AssertionGenerator.MUTATION_ANALYSIS
    """The generator that shall be used for assertion generation."""

    allow_stale_assertions: bool = False
    """Allow assertion on things that did not change between statement executions."""

    mutation_strategy: MutationStrategy = MutationStrategy.FIRST_ORDER_MUTANTS
    """The strategy that shall be used for creating mutants in the mutation analysis
    assertion generation method."""

    mutation_order: int = 1
    """The order of the generated higher order mutants in the mutation analysis
    assertion generation method."""

    post_process: bool = True
    """Should the results be post processed? For example, truncate test cases after
    statements that raise an exception."""

    float_precision: float = 0.01
    """Precision to use in float comparisons and assertions"""


# pylint:disable=too-many-instance-attributes
@dataclasses.dataclass
class SeedingConfiguration:
    """Configuration related to seeding."""

    seed: Optional[int] = None
    """A predefined seed value for the random number generator that is used."""

    constant_seeding: bool = True
    """Should the generator use a static constant seeding technique to improve constant
    generation?"""

    initial_population_seeding: bool = False
    """Should the generator use previously existing testcases to seed the initial
    population?"""

    initial_population_data: str = ""
    """The path to the file with the pre-existing tests. The path has to include the
    file itself."""

    seeded_testcases_reuse_probability: float = 0.9
    """Probability of using seeded testcases when initial population seeding is
    enabled."""

    initial_population_mutations: int = 0
    """Number of how often the testcases collected by initial population seeding should
    be mutated to promote diversity"""

    dynamic_constant_seeding: bool = True
    """Enables seeding of constants at runtime."""

    seeded_primitives_reuse_probability: float = 0.2
    """Probability for using seeded primitive values instead of randomly
    generated ones."""

    seeded_dynamic_values_reuse_probability: float = 0.6
    """Probability of using dynamically seeded values when a primitive seeded
     value will be used."""

    seed_from_archive: bool = False
    """When sampling new test cases reuse some from the archive, if one is used."""

    seed_from_archive_probability: float = 0.2
    """Instead of creating a new test case, reuse a covering solution from the archive,
    iff an archive is used."""

    seed_from_archive_mutations: int = 3
    """Number of mutations applied when sampling from the archive."""


@dataclasses.dataclass
class MIOPhaseConfiguration:
    """Configuration for a phase of MIO."""

    number_of_tests_per_target: int
    """Number of test cases for each target goal to keep in an archive."""

    random_test_or_from_archive_probability: float
    """Probability [0,1] of sampling a new test at random or choose an existing one in
    an archive."""

    number_of_mutations: int
    """Number of mutations allowed to be done on the same individual before
    sampling a new one."""


@dataclasses.dataclass
class MIOConfiguration:
    """Configuration that is specific to the MIO approach."""

    initial_config: MIOPhaseConfiguration = dataclasses.field(
        default_factory=lambda: MIOPhaseConfiguration(
            number_of_tests_per_target=10,
            random_test_or_from_archive_probability=0.5,
            number_of_mutations=1,
        )
    )
    """Configuration to use before focused phase."""

    focused_config: MIOPhaseConfiguration = dataclasses.field(
        default_factory=lambda: MIOPhaseConfiguration(
            number_of_tests_per_target=1,
            random_test_or_from_archive_probability=0.0,
            number_of_mutations=10,
        )
    )
    """Configuration to use in focused phase"""

    exploitation_starts_at_percent: float = 0.5
    """Percentage ]0,1] of search budget after which exploitation is activated, i.e.,
    switching to focused phase."""


@dataclasses.dataclass
class RandomConfiguration:
    """Configuration that is specific to the RANDOM approach."""

    max_sequence_length: int = 10
    """The maximum length of sequences that are generated, 0 means infinite."""

    max_sequences_combined: int = 10
    """The maximum number of combined sequences, 0 means infinite."""


@dataclasses.dataclass
class TypeInferenceConfiguration:
    """Configuration related to type inference."""

    guess_unknown_types: bool = True
    """Should we guess unknown types while constructing parameters?
    This might happen in the following cases:
    The parameter type is unknown, e.g. a parameter is missing a type hint.
    The parameter is not primitive and cannot be created from the test cluster,
    e.g. Callable[...]"""

    type_inference_strategy: TypeInferenceStrategy = TypeInferenceStrategy.TYPE_HINTS
    """The strategy for type-inference that shall be used"""

    max_cluster_recursion: int = 10
    """The maximum level of recursion when calculating the dependencies in the test
    cluster."""

    stub_dir: Optional[str] = None
    """Path to the pyi-stub files for the StubInferenceStrategy"""


@dataclasses.dataclass
class TestCreationConfiguration:
    """Configuration related to test creation."""

    max_recursion: int = 10
    """Recursion depth when trying to create objects in a test case."""

    max_delta: int = 20
    """Maximum size of delta for numbers during mutation"""

    max_int: int = 2048
    """Maximum size of randomly generated integers (minimum range = -1 * max)"""

    string_length: int = 20
    """Maximum length of randomly generated strings"""

    bytes_length: int = 20
    """Maximum length of randomly generated bytes"""

    collection_size: int = 5
    """Maximum length of randomly generated collections"""

    primitive_reuse_probability: float = 0.5
    """Probability to reuse an existing primitive in a test case, if available.
    Expects values in [0,1]"""

    object_reuse_probability: float = 0.9
    """Probability to reuse an existing object in a test case, if available.
    Expects values in [0,1]"""

    none_probability: float = 0.1
    """Probability to use None in a test case instead of constructing an object.
    Expects values in [0,1]"""

    skip_optional_parameter_probability: float = 0.7
    """Probability to skip an optional parameter, i.e., do not fill this parameter."""

    max_attempts: int = 1000
    """Number of attempts when generating an object before giving up"""

    insertion_uut: float = 0.5
    """Score for selection of insertion of UUT calls"""

    max_size: int = 100
    """Maximum number of test cases in a test suite"""


@dataclasses.dataclass
class SearchAlgorithmConfiguration:
    """General configuration for search algorithms."""

    min_initial_tests: int = 1
    """Minimum number of tests in initial test suites"""

    max_initial_tests: int = 10
    """Maximum number of tests in initial test suites"""

    population: int = 50
    """Population size of genetic algorithm"""

    chromosome_length: int = 40
    """Maximum length of chromosomes during search"""

    chop_max_length: bool = True
    """Chop statements after exception if length has reached maximum"""

    elite: int = 1
    """Elite size for search algorithm"""

    crossover_rate: float = 0.75
    """Probability of crossover"""

    test_insertion_probability: float = 0.1
    """Initial probability of inserting a new test in a test suite"""

    test_delete_probability: float = 1.0 / 3.0
    """Probability of deleting statements during mutation"""

    test_change_probability: float = 1.0 / 3.0
    """Probability of changing statements during mutation"""

    test_insert_probability: float = 1.0 / 3.0
    """Probability of inserting new statements during mutation"""

    statement_insertion_probability: float = 0.5
    """Initial probability of inserting a new statement in a test case"""

    random_perturbation: float = 0.2
    """Probability to replace a primitive with a random new value rather than adding
    a delta."""

    change_parameter_probability: float = 0.1
    """Probability of replacing parameters when mutating a method or constructor
    statement in a test case.  Expects values in [0,1]"""

    tournament_size: int = 5
    """Number of individuals for tournament selection."""

    rank_bias: float = 1.7
    """Bias for better individuals in rank selection"""

    selection: Selection = Selection.TOURNAMENT_SELECTION
    """The selection operator for genetic algorithms."""

    use_archive: bool = False
    """Some algorithms can be enhanced with an optional archive, e.g. Whole Suite ->
    Whole Suite + Archive. Use this option to enable the usage of an archive.
    Algorithms that always use an archive are not affected by this option."""

    filter_covered_targets_from_test_cluster: bool = False
    """Focus search by filtering out elements from the test cluster when
     they are fully covered."""


@dataclasses.dataclass
class StoppingConfiguration:
    """Configuration related to when Pynguin should stop."""

    stopping_condition: StoppingCondition = StoppingCondition.MAX_TIME
    """What condition should be checked to end the search/test generation."""

    budget: int = 600
    """Time budget (in seconds) that can be used for generating tests."""

    maximum_test_executions: int = 60
    """Maximum number of test cases to be executed."""

    maximum_statement_executions: int = 600
    """Maximum number of test cases to be executed."""

    maximum_iterations: int = 60
    """Maximum iterations"""


# pylint: disable=too-many-instance-attributes, pointless-string-statement
@dataclasses.dataclass
class Configuration:
    """General configuration for the test generator."""

    project_path: str
    """Path to the project the generator shall create tests for."""

    module_name: str
    """Name of the module for which the generator shall create tests."""

    test_case_output: TestCaseOutputConfiguration
    """Configuration for how test cases should be output."""

    algorithm: Algorithm = Algorithm.DYNAMOSA
    """The algorithm that shall be used for generation."""

    statistics_output: StatisticsOutputConfiguration = dataclasses.field(
        default_factory=StatisticsOutputConfiguration
    )
    """Statistic Output configuration."""

    stopping: StoppingConfiguration = dataclasses.field(
        default_factory=StoppingConfiguration
    )
    """Stopping configuration."""

    seeding: SeedingConfiguration = dataclasses.field(
        default_factory=SeedingConfiguration
    )
    """Seeding configuration."""

    type_inference: TypeInferenceConfiguration = dataclasses.field(
        default_factory=TypeInferenceConfiguration
    )
    """Type inference configuration."""

    test_creation: TestCreationConfiguration = dataclasses.field(
        default_factory=TestCreationConfiguration
    )
    """Test creation configuration."""

    search_algorithm: SearchAlgorithmConfiguration = dataclasses.field(
        default_factory=SearchAlgorithmConfiguration
    )
    """Search algorithm configuration."""

    mio: MIOConfiguration = dataclasses.field(default_factory=MIOConfiguration)
    """Configuration used for the MIO algorithm."""

    random: RandomConfiguration = dataclasses.field(default_factory=RandomConfiguration)
    """Configuration used for the RANDOM algorithm."""


# Singleton instance of the configuration.
configuration = Configuration(
    project_path="",
    module_name="",
    test_case_output=TestCaseOutputConfiguration(output_path=""),
)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an abstract base class for MOSA and its derivatives."""
from __future__ import annotations

import logging
from abc import ABCMeta
from typing import cast

import pynguin.configuration as config
import pynguin.ga.testcasechromosome as tcc
from pynguin.ga.comparators.dominancecomparator import DominanceComparator
from pynguin.generation.algorithms.archive import CoverageArchive
from pynguin.generation.algorithms.testgenerationstrategy import TestGenerationStrategy
from pynguin.utils import randomness
from pynguin.utils.exceptions import ConstructionFailedException


class AbstractMOSATestStrategy(
    TestGenerationStrategy[CoverageArchive], metaclass=ABCMeta
):
    """An abstract base implementation for MOSA and its derivatives."""

    _logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self._population: list[tcc.TestCaseChromosome] = []
        self._number_of_goals = -1

    def _breed_next_generation(self) -> list[tcc.TestCaseChromosome]:
        offspring_population: list[tcc.TestCaseChromosome] = []
        for _ in range(int(config.configuration.search_algorithm.population / 2)):
            parent_1 = self._selection_function.select(self._population)[0]
            parent_2 = self._selection_function.select(self._population)[0]
            offspring_1 = cast(tcc.TestCaseChromosome, parent_1.clone())
            offspring_2 = cast(tcc.TestCaseChromosome, parent_2.clone())

            # Apply crossover
            if (
                randomness.next_float()
                <= config.configuration.search_algorithm.crossover_rate
            ):
                try:
                    self._crossover_function.cross_over(offspring_1, offspring_2)
                except ConstructionFailedException:
                    self._logger.debug("CrossOver failed.")
                    continue

            # Apply mutation on offspring_1
            self._mutate(offspring_1)
            if offspring_1.has_changed() and offspring_1.size() > 0:
                offspring_population.append(offspring_1)

            # Apply mutation on offspring_2
            self._mutate(offspring_2)
            if offspring_2.has_changed() and offspring_2.size() > 0:
                offspring_population.append(offspring_2)

        # Add new randomly generated tests
        for _ in range(
            int(
                config.configuration.search_algorithm.population
                * config.configuration.search_algorithm.test_insertion_probability
            )
        ):
            if len(self._archive.covered_goals) == 0 or randomness.next_bool():
                tch: tcc.TestCaseChromosome = self._chromosome_factory.get_chromosome()
            else:
                tch = randomness.choice(self._archive.solutions).clone()
                tch.mutate()

            if tch.has_changed() and tch.size() > 0:
                offspring_population.append(tch)

        self._logger.debug("Number of offsprings = %d", len(offspring_population))
        return offspring_population

    @staticmethod
    def _mutate(offspring: tcc.TestCaseChromosome) -> None:
        offspring.mutate()
        if not offspring.has_changed():
            # if offspring is not changed, we try to mutate it once again
            offspring.mutate()

    def _get_non_dominated_solutions(
        self, solutions: list[tcc.TestCaseChromosome]
    ) -> list[tcc.TestCaseChromosome]:
        comparator: DominanceComparator[tcc.TestCaseChromosome] = DominanceComparator(
            goals=self._archive.covered_goals
        )
        next_front: list[tcc.TestCaseChromosome] = []
        for solution in solutions:
            is_dominated = False
            dominated_solutions: list[tcc.TestCaseChromosome] = []
            for best in next_front:
                flag = comparator.compare(solution, best)
                if flag < 0:
                    dominated_solutions.append(best)
                if flag > 0:
                    is_dominated = True
            if is_dominated:
                continue
            next_front.append(solution)
            for dominated_solution in dominated_solutions:
                if dominated_solution in next_front:
                    next_front.remove(dominated_solution)
        return next_front

    def _get_random_population(self) -> list[tcc.TestCaseChromosome]:
        population: list[tcc.TestCaseChromosome] = []
        for _ in range(config.configuration.search_algorithm.population):
            chromosome = self._chromosome_factory.get_chromosome()
            population.append(chromosome)
        return population

    def _get_best_individuals(self) -> list[tcc.TestCaseChromosome]:
        return self._get_non_dominated_solutions(self._population)



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an implementation to generate statements out of an AST."""
from __future__ import annotations

import ast
import inspect
import logging
from typing import TYPE_CHECKING, Any, cast

import pynguin.assertion.assertion as ass
import pynguin.testcase.defaulttestcase as dtc
import pynguin.testcase.statement as stmt
from pynguin.utils.generic.genericaccessibleobject import (
    GenericCallableAccessibleObject,
    GenericConstructor,
    GenericFunction,
    GenericMethod,
)
from pynguin.utils.type_utils import is_assertable

if TYPE_CHECKING:
    import pynguin.testcase.testcase as tc
    import pynguin.testcase.variablereference as vr
    from pynguin.setup.testcluster import TestCluster

logger = logging.getLogger(__name__)


def create_assign_stmt(
    assign: ast.Assign,
    testcase: tc.TestCase,
    ref_dict: dict[str, vr.VariableReference],
    test_cluster: TestCluster,
) -> tuple[str, stmt.VariableCreatingStatement] | None:
    """Creates the corresponding statement from an ast.Assign node.

    Args:
        assign: The ast.Assign node
        testcase: The testcase of the statement
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references.
        test_cluster: The test cluster that is used to resolve classes, methods, etc.

    Returns:
        The corresponding statement or None if no statement type matches.
    """
    new_stmt: stmt.VariableCreatingStatement | None
    value = assign.value
    objs_under_test = test_cluster.accessible_objects_under_test
    callable_objects_under_test: set[GenericCallableAccessibleObject] = {
        o for o in objs_under_test if isinstance(o, GenericCallableAccessibleObject)
    }
    if isinstance(value, ast.Constant):
        new_stmt = create_stmt_from_constant(value, testcase)
    elif isinstance(value, ast.UnaryOp):
        new_stmt = create_stmt_from_unaryop(value, testcase)
    elif isinstance(value, ast.Call):
        new_stmt = create_stmt_from_call(
            value, testcase, callable_objects_under_test, ref_dict
        )
    elif isinstance(value, (ast.List, ast.Set, ast.Dict, ast.Tuple)):
        new_stmt = create_stmt_from_collection(
            value, testcase, callable_objects_under_test, ref_dict
        )
    else:
        logger.info("Assign statement could not be parsed.")
        new_stmt = None
    if new_stmt is None:
        return None
    ref_id = str(assign.targets[0].id)  # type: ignore
    return ref_id, new_stmt


def create_assert_stmt(
    ref_dict: dict[str, vr.VariableReference], assert_node: ast.Assert
) -> tuple[ass.Assertion, vr.VariableReference] | None:
    """Creates an assert statement.

    Args:
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references.
        assert_node: the ast assert node.

    Returns:
        The corresponding assert statement.
    """
    assertion: ass.Assertion | None = None
    try:
        source = ref_dict[assert_node.test.left.id]  # type: ignore
        val_elem = assert_node.test.comparators[0]  # type: ignore
        operator = assert_node.test.ops[0]  # type: ignore
    except (KeyError, AttributeError):
        return None
    if isinstance(operator, (ast.Is, ast.Eq)):
        assertion = create_assertion(source, val_elem)
    if assertion is not None:
        return assertion, source
    return None


def create_assertion(
    source: vr.VariableReference,
    val_elem: ast.Constant | ast.UnaryOp | None,
) -> ass.Assertion | None:
    """Creates an assertion.

    Args:
        source: The variable reference
        val_elem: The ast element for retrieving the value

    Returns:
        The assertion.
    """
    if isinstance(val_elem, ast.UnaryOp):
        val_elem = val_elem.operand  # type: ignore

    if isinstance(val_elem, ast.Constant) and is_assertable(val_elem.value):
        return ass.ObjectAssertion(source, val_elem.value)
    return None


def create_variable_references_from_call_args(
    call_args: list[ast.Name | ast.Starred],
    call_keywords: list[ast.keyword],
    gen_callable: GenericCallableAccessibleObject,
    ref_dict: dict[str, vr.VariableReference],
) -> dict[str, vr.VariableReference] | None:
    """Takes the arguments of an ast.Call node and returns the variable references of
    the corresponding statements.

    Args:
        call_args: the positional arguments
        call_keywords: the keyword arguments
        gen_callable: the callable that is called
        ref_dict: a dictionary containing the variable references

    Returns:
        The dict with the variable references of the call_args.

    """
    var_refs: dict[str, vr.VariableReference] = {}
    # We have to ignore the first parameter (usually 'self') for regular methods and
    # constructors because it is filled by the runtime.
    # TODO(fk) also consider @classmethod, because their first argument is the class,
    #  which is also filled by the runtime.
    shift_by = 1 if gen_callable.is_method() or gen_callable.is_constructor() else 0

    # Handle positional arguments.
    for (name, param), call_arg in zip(
        list(gen_callable.inferred_signature.signature.parameters.items())[shift_by:],
        call_args,
    ):
        if (
            param.kind
            in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        ) and isinstance(call_arg, ast.Name):
            reference = ref_dict.get(call_arg.id)
        elif param.kind == inspect.Parameter.VAR_POSITIONAL and isinstance(
            call_arg, ast.Starred
        ):
            reference = ref_dict.get(call_arg.value.id)  # type: ignore
        else:
            return None
        if reference is None:
            # Reference could not be resolved
            return None
        var_refs[name] = reference

    # Handle keyword arguments
    for call_keyword in call_keywords:
        keyword = call_keyword.arg
        if keyword is None:
            # **kwargs has to be the last parameter?
            keyword = list(gen_callable.inferred_signature.signature.parameters.keys())[
                -1
            ]
            if (
                gen_callable.inferred_signature.signature.parameters[keyword].kind
                != inspect.Parameter.VAR_KEYWORD
            ):
                return None
        if not isinstance(call_keyword.value, ast.Name):
            return None
        reference = ref_dict.get(call_keyword.value.id)
        if reference is None:
            return None
        var_refs[keyword] = reference

    return var_refs


# pylint: disable=too-many-return-statements
def create_stmt_from_constant(
    constant: ast.Constant, testcase: tc.TestCase
) -> stmt.VariableCreatingStatement | None:
    """Creates a statement from an ast.constant node.

    Args:
        constant: the ast.Constant statement
        testcase: the testcase containing the statement

    Returns:
        The corresponding statement.
    """
    if constant.value is None:
        return stmt.NoneStatement(testcase, constant.value)

    val = constant.value
    if isinstance(val, bool):
        return stmt.BooleanPrimitiveStatement(testcase, val)
    if isinstance(val, int):
        return stmt.IntPrimitiveStatement(testcase, val)
    if isinstance(val, float):
        return stmt.FloatPrimitiveStatement(testcase, val)
    if isinstance(val, str):
        return stmt.StringPrimitiveStatement(testcase, val)
    if isinstance(val, bytes):
        return stmt.BytesPrimitiveStatement(testcase, val)
    logger.info("Could not find case for constant while handling assign statement.")
    return None


def create_stmt_from_unaryop(
    unaryop: ast.UnaryOp, testcase: tc.TestCase
) -> stmt.VariableCreatingStatement | None:
    """Creates a statement from an ast.unaryop node.

    Args:
        unaryop: the ast.UnaryOp statement
        testcase: the testcase containing the statement

    Returns:
        The corresponding statement.
    """
    val = unaryop.operand.value  # type: ignore
    if isinstance(val, bool):
        return stmt.BooleanPrimitiveStatement(testcase, not val)
    if isinstance(val, float):
        return stmt.FloatPrimitiveStatement(testcase, (-1) * val)
    if isinstance(val, int):
        return stmt.IntPrimitiveStatement(testcase, (-1) * val)
    logger.info(
        "Could not find case for unary operator while handling assign statement."
    )
    return None


def create_stmt_from_call(
    call: ast.Call,
    testcase: tc.TestCase,
    objs_under_test: set[GenericCallableAccessibleObject],
    ref_dict: dict[str, vr.VariableReference],
) -> stmt.VariableCreatingStatement | None:
    """Creates the corresponding statement from an ast.call node. Depending on the call,
    this can be a GenericConstructor, GenericMethod or GenericFunction statement.

    Args:
        call: the ast.Call node
        testcase: the testcase of the statement
        objs_under_test: the accessible objects under test
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references.

    Returns:
        The corresponding statement.
    """
    try:
        call.func.attr  # type: ignore
    except AttributeError:
        return try_generating_specific_function(
            call, testcase, objs_under_test, ref_dict
        )
    gen_callable = find_gen_callable(call, objs_under_test, ref_dict)
    if gen_callable is None:
        logger.info("No such function found...")
        return None
    return assemble_stmt_from_gen_callable(testcase, gen_callable, call, ref_dict)


def find_gen_callable(
    call: ast.Call,
    objs_under_test: set,
    ref_dict: dict[str, vr.VariableReference],
) -> GenericConstructor | GenericMethod | GenericFunction | None:
    """Traverses the accessible objects under test and returns the one matching with the
    ast.call object. Unfortunately, there is no possibility to clearly determine if the
    ast.call object is a constructor, method or function. Hence, the looping over all
    accessible objects is unavoidable. Then, by the name of the ast.call and by the
    owner (functions do not have one, constructors and methods have), it is possible to
    decide which accessible object to choose. This should also be unique, because the
    name of a function should be unique in a module. The name of a method should be
    unique inside one class. If two classes in the same module have a method with an
    equal name, the right method can be determined by the type of the object that is
    calling the method. This object has the type of the class of which the method is
    called. To determine between function names and method names, another thing needs
    to be considered. If a method is called, it is called on an object. This object must
    have been created before the function is called on that object. Thus, this object
    must have been initialized before and have a variable reference in the ref_dict
    where all created variable references are stored. So, by checking, if a reference is
    found, it can be decided if it is a function or a method.

        Args:
            call: the ast.Call node
            objs_under_test: the accessible objects under test
            ref_dict: a dictionary containing key value pairs of variable ids and
                      variable references.

        Returns:
            The corresponding generic accessible object under test. This can be a
            GenericConstructor, a GenericMethod or a GenericFunction.
    """
    call_name = str(call.func.attr)  # type: ignore
    for obj in objs_under_test:
        if isinstance(obj, GenericConstructor):
            owner = str(obj.owner).rsplit(".", maxsplit=1)[-1].split("'")[0]
            call_id = call.func.value.id  # type: ignore
            if call_name == owner and call_id not in ref_dict:
                return obj
        elif isinstance(obj, GenericMethod):
            # test if the type of the calling object is equal to the type of the owner
            # of the generic method
            call_id = call.func.value.id  # type: ignore
            if call_name == obj.method_name and call_id in ref_dict:
                obj_from_ast = str(call.func.value.id)  # type: ignore
                var_type = ref_dict[obj_from_ast].type
                if var_type == obj.owner:
                    return obj
        elif isinstance(obj, GenericFunction):
            if call_name == obj.function_name:
                return obj
    return None


def assemble_stmt_from_gen_callable(
    testcase: tc.TestCase,
    gen_callable: GenericCallableAccessibleObject,
    call: ast.Call,
    ref_dict: dict[str, vr.VariableReference],
) -> stmt.ParametrizedStatement | None:
    """Takes a generic callable and assembles the corresponding parametrized statement
    from it.

    Args:
        testcase: the testcase of the statement
        gen_callable: the corresponding callable of the cluster
        call: the ast.Call statement
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references.

    Returns:
        The corresponding statement.
    """
    for arg in call.args:
        if not isinstance(arg, (ast.Name, ast.Starred)):
            return None
    for keyword in call.keywords:
        if not isinstance(keyword, ast.keyword):
            return None
    var_refs = create_variable_references_from_call_args(
        call.args, call.keywords, gen_callable, ref_dict  # type: ignore
    )
    if var_refs is None:
        return None
    if isinstance(gen_callable, GenericFunction):
        return stmt.FunctionStatement(
            testcase, cast(GenericCallableAccessibleObject, gen_callable), var_refs
        )
    if isinstance(gen_callable, GenericMethod):
        return stmt.MethodStatement(
            testcase,
            gen_callable,
            ref_dict[call.func.value.id],  # type: ignore
            var_refs,
        )
    if isinstance(gen_callable, GenericConstructor):
        return stmt.ConstructorStatement(
            testcase, cast(GenericCallableAccessibleObject, gen_callable), var_refs
        )
    return None


def create_stmt_from_collection(
    coll_node: ast.List | ast.Set | ast.Dict | ast.Tuple,
    testcase: tc.TestCase,
    objs_under_test: set[GenericCallableAccessibleObject],
    ref_dict: dict[str, vr.VariableReference],
) -> stmt.VariableCreatingStatement | None:
    """Creates the corresponding statement from an ast.List node. Lists contain other
    statements.

    Args:
        coll_node: the ast node. It has the type of one of the collection types.
        testcase: the testcase of the statement
        objs_under_test: the accessible objects under test. Not needed for the
        collection statement, but lists can contain other statements (e.g. call) needing
        this.
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references. Not needed for the collection statement, but
                  lists can contain other statements (e.g. call) needing this.

    Returns:
        The corresponding list statement.
    """
    coll_elems: None | (
        list[vr.VariableReference]
        | list[tuple[vr.VariableReference, vr.VariableReference]]
    )
    if isinstance(coll_node, ast.Dict):
        keys = create_elements(coll_node.keys, testcase, objs_under_test, ref_dict)
        values = create_elements(coll_node.values, testcase, objs_under_test, ref_dict)
        if keys is None or values is None:
            return None
        coll_elems_type = get_collection_type(values)
        coll_elems = list(zip(keys, values))
    else:
        elements = coll_node.elts
        coll_elems = create_elements(elements, testcase, objs_under_test, ref_dict)
        if coll_elems is None:
            return None
        coll_elems_type = get_collection_type(coll_elems)
    return create_specific_collection_stmt(
        testcase, coll_node, coll_elems_type, coll_elems
    )


def create_elements(
    elements: Any,
    testcase: tc.TestCase,
    objs_under_test: set[GenericCallableAccessibleObject],
    ref_dict: dict[str, vr.VariableReference],
) -> list[vr.VariableReference] | None:
    """Creates the elements of a collection by calling the corresponding methods for
    creation. This can be recursive.

    Args:
        elements: The elements of the collection
        testcase: the corresponding testcase
        objs_under_test: A set of generic accessible objects under test
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references

    Returns:
        A list of variable references or None if something goes wrong while creating the
        elements.
    """
    coll_elems: list[vr.VariableReference] = []
    for elem in elements:
        statement: stmt.VariableCreatingStatement | None
        if isinstance(elem, ast.Constant):
            statement = create_stmt_from_constant(elem, testcase)
            if not statement:
                return None
            coll_elems.append(testcase.add_variable_creating_statement(statement))
        elif isinstance(elem, ast.UnaryOp):
            statement = create_stmt_from_unaryop(elem, testcase)
            if not statement:
                return None
            coll_elems.append(testcase.add_variable_creating_statement(statement))
        elif isinstance(elem, ast.Call):
            statement = create_stmt_from_call(elem, testcase, objs_under_test, ref_dict)
            if not statement:
                return None
            coll_elems.append(testcase.add_variable_creating_statement(statement))
        elif isinstance(elem, (ast.List, ast.Tuple, ast.Set, ast.Dict)):
            statement = create_stmt_from_collection(
                elem, testcase, objs_under_test, ref_dict
            )
            if not statement:
                return None
            coll_elems.append(testcase.add_variable_creating_statement(statement))
        elif isinstance(elem, ast.Name):
            try:
                coll_elems.append(ref_dict[elem.id])
            except AttributeError:
                return None
        else:
            return None
    return coll_elems


def get_collection_type(coll_elems: list[vr.VariableReference]) -> Any:
    """Returns the type of a collection. If objects of multiple types are in the
    collection, this function returns None.

    Args:
        coll_elems: a list of variable references

    Returns:
        The type of the collection.
    """
    if len(coll_elems) == 0:
        return None
    coll_type = coll_elems[0].type
    for elem in coll_elems:
        if not elem.type == coll_type:
            coll_type = None
            break
    return coll_type


def create_specific_collection_stmt(
    testcase: tc.TestCase,
    coll_node: ast.List | ast.Set | ast.Dict | ast.Tuple,
    coll_elems_type: Any,
    coll_elems: list[Any],
) -> None | (
    stmt.ListStatement | stmt.SetStatement | stmt.DictStatement | stmt.TupleStatement
):
    """Creates the corresponding collection statement from an ast node.

    Args:
        testcase: The testcase of the statement
        coll_node: the ast node
        coll_elems: a list of variable references or a list of tuples of variables for a
        dict statement.
        coll_elems_type: the type of the elements of the collection statement.

    Returns:
        The corresponding collection statement.
    """
    if isinstance(coll_node, ast.List):
        return stmt.ListStatement(testcase, coll_elems_type, coll_elems)
    if isinstance(coll_node, ast.Set):
        return stmt.SetStatement(testcase, coll_elems_type, coll_elems)
    if isinstance(coll_node, ast.Dict):
        return stmt.DictStatement(testcase, coll_elems_type, coll_elems)
    if isinstance(coll_node, ast.Tuple):
        return stmt.TupleStatement(testcase, coll_elems_type, coll_elems)
    return None


def try_generating_specific_function(
    call: ast.Call,
    testcase: tc.TestCase,
    objs_under_test: set[GenericCallableAccessibleObject],
    ref_dict: dict[str, vr.VariableReference],
) -> stmt.VariableCreatingStatement | None:
    """Calls to creating a collection (list, set, tuple, dict) via their keywords and
    not via literal syntax are considered as ast.Call statements. But for these calls,
    no accessible object under test is in the test_cluster. To parse them anyway, these
    method transforms them to the corresponding ast statement, for example a call of a
    list with 'list()' to an ast.List statement.

    Args:
        call: the ast.Call node
        testcase: the testcase of the statement
        objs_under_test: the accessible objects under test
        ref_dict: a dictionary containing key value pairs of variable ids and
                  variable references.

    Returns:
        The corresponding statement.

    """
    try:
        func_id = str(call.func.id)  # type: ignore
    except AttributeError:
        return None
    if func_id == "set":
        try:
            set_node = ast.Set(
                elts=call.args,
                ctx=ast.Load(),
            )
        except AttributeError:
            return None
        return create_stmt_from_collection(
            set_node, testcase, objs_under_test, ref_dict
        )
    if func_id == "list":
        try:
            list_node = ast.List(
                elts=call.args,
                ctx=ast.Load(),
            )
        except AttributeError:
            return None
        return create_stmt_from_collection(
            list_node, testcase, objs_under_test, ref_dict
        )
    if func_id == "tuple":
        try:
            tuple_node = ast.Tuple(
                elts=call.args,
                ctx=ast.Load(),
            )
        except AttributeError:
            return None
        return create_stmt_from_collection(
            tuple_node, testcase, objs_under_test, ref_dict
        )
    if func_id == "dict":
        try:
            dict_node = ast.Dict(
                keys=call.args[0].keys if call.args else [],  # type: ignore
                values=call.args[0].values if call.args else [],  # type: ignore
                ctx=ast.Load(),
            )
        except AttributeError:
            return None
        return create_stmt_from_collection(
            dict_node, testcase, objs_under_test, ref_dict
        )
    return None


# pylint: disable=invalid-name, missing-function-docstring
class AstToTestCaseTransformer(ast.NodeVisitor):
    """A AST NodeVisitor that tries to convert an AST into our internal
    test case representation."""

    def __init__(self, test_cluster: TestCluster, create_assertions: bool):
        self._current_testcase: dtc.DefaultTestCase = dtc.DefaultTestCase()
        self._current_parsable: bool = True
        self._var_refs: dict[str, vr.VariableReference] = {}
        self._testcases: list[dtc.DefaultTestCase] = []
        self._number_found_testcases: int = 0
        self._test_cluster = test_cluster
        self._create_assertions = create_assertions

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        self._number_found_testcases += 1
        self._current_testcase = dtc.DefaultTestCase()
        self._current_parsable = True
        self._var_refs.clear()
        self.generic_visit(node)
        if self._current_parsable:
            self._testcases.append(self._current_testcase)

    def visit_Assign(self, node: ast.Assign) -> Any:
        if self._current_parsable:
            if (
                result := create_assign_stmt(
                    node, self._current_testcase, self._var_refs, self._test_cluster
                )
            ) is None:
                self._current_parsable = False
            else:
                ref_id, stm = result
                var_ref = self._current_testcase.add_variable_creating_statement(stm)
                self._var_refs[ref_id] = var_ref

    def visit_Assert(self, node: ast.Assert) -> Any:
        if self._current_parsable and self._create_assertions:
            if (result := create_assert_stmt(self._var_refs, node)) is not None:
                assertion, var_ref = result
                self._current_testcase.get_statement(
                    var_ref.get_statement_position()
                ).add_assertion(assertion)

    @property
    def testcases(self) -> list[dtc.DefaultTestCase]:
        """Provides the testcases that could be generated from the given AST.
        It is possible that not every aspect of the AST could be transformed
        to our internal representation.

        Returns:
            The generated testcases.
        """
        return self._testcases



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
"""Provides tools to build an inheritance graph of the subject under test."""
from __future__ import annotations

import inspect
from typing import Any, NamedTuple, overload

import networkx as nx


class ClassInformation(NamedTuple):
    """Stores class information in form of fully-qualified name and class object."""

    name: str
    class_object: type

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, ClassInformation):
            return False
        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return f"ClassInformation({self.name}, {self.class_object})"

    def __repr__(self) -> str:
        return f"ClassInformation({self.name}, {self.class_object})"


class InheritanceGraph:
    """Represents the inheritance structure of the found types."""

    def __init__(self) -> None:
        self._graph = nx.DiGraph()

    def add_edge(self, source: ClassInformation, target: ClassInformation) -> None:
        """Adds an edge to the graph.

        An edge is a sub-type relation between two classes, as yielded, for example,
        by querying the method-resolution order.

        Args:
            source: The super class
            target: The sub class
        """
        self._graph.add_edge(source, target)

    def find(self, class_information: ClassInformation) -> ClassInformation | None:
        """Finds a class in the inheritance graph.

        Args:
            class_information: The class to search for

        Returns:
            The class, if it was found.  None otherwise.
        """
        if self._graph.has_node(class_information):
            for node in self._graph.nodes.keys():
                if class_information == node:
                    return node
        return None

    def get_sub_types(
        self, class_information: ClassInformation
    ) -> set[ClassInformation]:
        """Provides a set of sub types for a class.

        If there are no sub types, the returned set is empty.

        Args:
            class_information: The class to start

        Returns:
            The sub types of the class
        """
        if graph_node := self.find(class_information):
            return self._get_transitive_successors(graph_node, set())
        raise ValueError("Node for class not found in inheritance graph.")

    def _get_transitive_successors(
        self, node: ClassInformation, done: set[ClassInformation]
    ) -> set[ClassInformation]:
        successors: set[ClassInformation] = set()
        for successor_node in self._get_successors(node):
            if successor_node not in done:
                successors.add(successor_node)
                done.add(successor_node)
                successors.update(self._get_transitive_successors(successor_node, done))
        return successors

    def _get_successors(self, node: ClassInformation) -> set[ClassInformation]:
        successors: set[ClassInformation] = set()
        for successor in self._graph.successors(node):
            successors.add(successor)
        return successors

    def get_super_types(
        self, class_information: ClassInformation
    ) -> set[ClassInformation]:
        """Provides a set of super types for a class.

        If there are no super types, the returned set is empty.

        Args:
            class_information: The class to start

        Returns:
            The super types of the class
        """
        if graph_node := self.find(class_information):
            return self._get_transitive_predecessors(graph_node, set())
        raise ValueError("Node for class not found in inheritance graph.")

    def _get_transitive_predecessors(
        self, node: ClassInformation, done: set[ClassInformation]
    ) -> set[ClassInformation]:
        predecessors: set[ClassInformation] = set()
        for predecessor_node in self._get_predecessors(node):
            if predecessor_node not in done:
                predecessors.add(predecessor_node)
                done.add(predecessor_node)
                predecessors.update(
                    self._get_transitive_predecessors(predecessor_node, done)
                )
        return predecessors

    def _get_predecessors(self, node: ClassInformation) -> set[ClassInformation]:
        predecessors: set[ClassInformation] = set()
        for predecessor in self._graph.predecessors(node):
            predecessors.add(predecessor)
        return predecessors

    def get_distance(self, source: ClassInformation, target: ClassInformation) -> int:
        """Computes the shortest distance between two nodes of the graph.

        The distance will be positive if target is a sub-type of source and negative
        if target is a super-type of source.  If source and target are the same node,
        the distance will be 0.

        Args:
            source: The source node
            target: The target node

        Returns:
            The (shortest) distance between the two nodes in the graph

        Raises:
            ValueError: if one of the class-information objects are not part of the
            graph
            nx.NetworkXNoPath: if no path between the nodes exists
        """
        if self.find(source) is None or self.find(target) is None:
            raise ValueError("Both elements need to be part of the graph")
        try:
            return nx.shortest_path_length(self._graph, source, target)
        except nx.NetworkXNoPath:
            return (-1) * nx.shortest_path_length(self._graph, target, source)

    def number_of_nodes(self) -> int:
        """Provides the number of nodes in the graph.

        Returns:
            The number of nodes in the graph
        """
        return self._graph.number_of_nodes()

    def number_of_edges(self) -> int:
        """Provides the number of edges in the graph.

        Returns:
            The number of edges in the graph
        """
        return self._graph.number_of_edges()


def build_inheritance_graph(analysed_classes: set[type]) -> InheritanceGraph:
    """Builds an inheritance graph for a set of types.

    Args:
        analysed_classes: The set of available types in the program

    Returns:
        The inheritance graph for these types
    """
    graph = InheritanceGraph()
    for analysed_class in analysed_classes:
        mro_class_information = [
            _retrieve_class_information(class_)
            for class_ in inspect.getmro(analysed_class)
        ]
        for target, source in zip(mro_class_information, mro_class_information[1:]):
            graph.add_edge(source, target)
    return graph


@overload
def build_class_information(value: ClassInformation) -> ClassInformation:
    ...  # pragma: no cover


@overload
def build_class_information(value: str) -> ClassInformation:
    ...  # pragma: no cover


@overload
def build_class_information(value: type) -> ClassInformation:
    ...  # pragma: no cover


def build_class_information(value) -> ClassInformation:
    """Builds the class-information object.

    The parameter can either be of type `ClassInformation`, `str`, or `type`.  For
    other inputs, a `ValueError` is raised.

    Args:
        value: The value to build the class-information object from

    Returns:
        A class-information object

    Raises:
        ValueError: In case no appropriate input was given
    """
    if isinstance(value, ClassInformation):
        return value
    if isinstance(value, str):
        return ClassInformation(name=value, class_object=type(None))
    if isinstance(value, type):
        return _retrieve_class_information(value)
    raise ValueError(f"Cannot build ClassInformation for {value}")


def _retrieve_class_information(class_object: type) -> ClassInformation:
    qualified_name = class_object.__qualname__
    name = class_object.__name__
    module_name = class_object.__module__
    if name == qualified_name:
        qualified_name = f"{module_name}.{name}"
    class_information = ClassInformation(name=qualified_name, class_object=class_object)
    return class_information



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides an output trace."""
from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from ordered_set import OrderedSet

if TYPE_CHECKING:
    import pynguin.assertion.assertion as ass
    import pynguin.testcase.statement as stmt


class AssertionTrace:
    """Stores the states that have been observed during execution."""

    def __init__(self) -> None:
        """Create new assertion trace."""
        # One entry per statement, i.e., the assertions after executing that statement.
        self._trace: dict[int, OrderedSet[ass.Assertion]] = defaultdict(OrderedSet)

    def add_entry(self, position: int, assertion: ass.Assertion) -> None:
        """Add an entry to this trace.

        Args:
            position: the position of the statement after whose execution the
                state is observed.
            assertion: the made assertion.
        """
        self._trace[position].add(assertion)

    def get_assertions(self, statement: stmt.Statement) -> OrderedSet[ass.Assertion]:
        """Get all assertions contained within this trace for the given statement.

        Args:
            statement: the statement for which all recorded assertions
                should be generated.

        Returns:
            All assertions in this trace for the given statement.
        """
        position = statement.get_position()
        if position in self._trace:
            return OrderedSet(self._trace[position])
        return OrderedSet()

    def clear(self) -> None:
        """Clear this trace."""
        self._trace.clear()

    def clone(self) -> AssertionTrace:
        """Clone this trace.

        Returns:
            a clone of this trace.
        """
        copy: AssertionTrace = AssertionTrace()
        for stmt_key, stmt_value in self._trace.items():
            copy._trace[stmt_key] = OrderedSet()
            for entry in stmt_value:
                copy._trace[stmt_key].add(entry)
        return copy



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""
Provides an atomic integer implementation similar to the Java class
`AtomicInteger`.
"""
import threading


class AtomicInteger:
    """An atomic integer implementation.

    This class is thread-safe, it does not rely on the global interpreter lock.

    Adapted from https://stackoverflow.com/a/48433648
    """

    def __init__(self, value: int = 0) -> None:
        self._value = value
        self._lock = threading.Lock()

    def inc(self) -> int:
        """Increments the value of the integer by one and returns the new value.

        Returns:
            The new value of the atomic integer
        """
        with self._lock:
            self._value += 1
            return self._value

    def dec(self) -> int:
        """Decrements the value of the integer by one and returns the new value.

        Returns:
            The new value of the atomic integer
        """
        with self._lock:
            self._value -= 1
            return self._value

    @property
    def value(self) -> int:
        """Provides the current value of the atomic integer.

        Returns:
            The current value of the atomic integer
        """
        with self._lock:
            return self._value

    @value.setter
    def value(self, value: int) -> None:
        """Sets the current value of the atomic integer and returns it.

        Args:
            value: The new value for the atomic integer
        """
        with self._lock:
            self._value = value



#  This file is part of Pynguin.
#
#  SPDX-FileCopyrightText: 2019–2022 Pynguin Contributors
#
#  SPDX-License-Identifier: LGPL-3.0-or-later
#
"""Provides archives to store found solutions."""
from __future__ import annotations

import logging
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Iterable

from ordered_set import OrderedSet

import pynguin.ga.computations as ff
import pynguin.ga.testcasechromosome as tcc
from pynguin.utils import randomness


class Archive(ABC):
    """Abstract base class for archives that store individuals."""

    _logger = logging.getLogger(__name__)

    def __init__(self):
        self._on_target_covered_callbacks: list[
            Callable[[ff.TestCaseFitnessFunction], None]
        ] = []

    @abstractmethod
    def update(self, solutions: Iterable[tcc.TestCaseChromosome]) -> bool:
        """Updates this archive with the given set of solutions.

        In detail, when a solution manages to satisfy a previously uncovered target,
        it is stored in the archive.  If a target is already covered by one solution
        in the archive, and the given list contains another solution that covers the
        same target, the shorter of the two solutions is retained.  Otherwise,
        the solution is discarded.

        Args:
            solutions: The solutions to update the archive with

        Returns:
            True, iff a solution was stored.
        """

    @property
    @abstractmethod
    def solutions(self) -> OrderedSet[tcc.TestCaseChromosome]:
        """Provides the best solutions found so far.

        Best solutions are those shortest test cases covering one of the targets
        managed by this archive.  Duplicates are eliminated.

        Returns:
            The best solutions in the archive
        """

    def add_on_target_covered(
        self, callback: Callable[[ff.TestCaseFitnessFunction], None]
    ) -> None:
        """Register a callback that is called once when a new target is covered for the
        first time.

        Args:
            callback: The call back to be registered.
        """
        self._on_target_covered_callbacks.append(callback)

    def _on_target_covered(self, target: ff.TestCaseFitnessFunction) -> None:
        self._logger.debug("Target covered: %s", target)
        for callback in self._on_target_covered_callbacks:
            callback(target)


class CoverageArchive(Archive):
    """Implements the archive used by MOSA.

    The archive is used to store the shortest covering test cases found so far.  It
    also keeps track of which targets are yet to cover.

    For each objective of MOSA, the archive stores the chromosome which covers that
    objective.  Furthermore, when two chromosomes cover the same objective,
    the archive retains the smaller of the two, thus automatically optimising for the
    test case length as secondary criterion.

    It is imperative to know the best possible fitness value for a given fitness
    function.  Otherwise, the archive cannot determine whether or not the target has
    been reached.  Therefore, we limit ourselves to minimising fitness functions and
    assume 0.0 the best possible value.  This does not restrict generality as it is
    trivial to convert from maximising to minimising fitness functions.
    """

    _logger = logging.getLogger(__name__)

    def __init__(self, objectives: OrderedSet[ff.TestCaseFitnessFunction]) -> None:
        super().__init__()
        self._covered: dict[ff.TestCaseFitnessFunction, tcc.TestCaseChromosome] = {}
        self._uncovered = OrderedSet(objectives)
        self._objectives = OrderedSet(objectives)

    def update(self, solutions: Iterable[tcc.TestCaseChromosome]) -> bool:
        """Updates this archive with the given set of solutions.

        In detail, when a solution manages to satisfy a previously uncovered target,
        it is stored in the archive.  If a target is already covered by one solution
        in the archive, and the given list contains another solution that covers the
        same target, the shorter of the two solutions is retained.  Otherwise,
        the solution is discarded.

        Args:
            solutions: The solutions to update the archive with
        """
        updated = False
        for objective in self._objectives:
            best_solution = self._covered.get(objective, None)
            best_size = sys.maxsize if best_solution is None else best_solution.size()

            for solution in solutions:
                covers = solution.get_is_covered(objective)
                size = solution.size()

                if covers and size < best_size:
                    updated = True
                    self._covered[objective] = solution
                    best_size = size
                    if objective in self._uncovered:
                        self._uncovered.remove(objective)
                        self._on_target_covered(objective)
        self._logger.debug("ArchiveCoverageGoals: %d", len(self._covered))
        return updated

    @property
    def uncovered_goals(self) -> OrderedSet[ff.TestCaseFitnessFunction]:
        """Provides the set of goals that are yet to cover.

        Returns:
            The uncovered goals
        """
        return self._uncovered

    @property
    def covered_goals(self) -> OrderedSet[ff.TestCaseFitnessFunction]:
        """Provides the set of goals that are already covered.

        Returns:
            The covered goals
        """
        return OrderedSet(self._covered.keys())

    @property
    def objectives(self) -> OrderedSet[ff.TestCaseFitnessFunction]:
        """Provides the set of all objectives.

        Returns:
            All objectives
        """
        return self._objectives

    def add_goals(self, new_goals: OrderedSet[ff.TestCaseFitnessFunction]) -> None:
        """Add goals to the archive to consider."""
        for goal in new_goals:
            if goal not in self._objectives:
                self._logger.debug("Adding goal: %s", goal)
                self._objectives.add(goal)
                self._uncovered.add(goal)

    @property
    def solutions(self) -> OrderedSet[tcc.TestCaseChromosome]:
        assert self._all_covered(), "Some covered targets have a fitness != 0.0"
        return OrderedSet(self._covered.values())

    def reset(self) -> None:
        """Resets the archive."""
        self._uncovered.update(self._objectives)
        self._covered.clear()

    def _all_covered(self) -> bool:
        return all(
            chromosome.get_is_covered(fitness_function)
            for fitness_function, chromosome in self._covered.items()
        )


@dataclass(frozen=True)
class MIOPopulationPair:
    """A tuple of h-value and corresponding test case chromosome."""

    # The h-value as described in the MIO paper.
    # pylint: disable=invalid-name
    h: float

    # The test case chromosome.
    test_case_chromosome: tcc.TestCaseChromosome


class MIOPopulation:
    """The population that is stored per target."""

    def __init__(self, population_size: int) -> None:
        self._counter = 0
        self._capacity = population_size
        # Assumption: These are always sorted.
        self._solutions: list[MIOPopulationPair] = []

    @property
    def counter(self) -> int:
        """How often was a solution sampled from this population since the last
        improvement in the h-value."""
        return self._counter

    @property
    def is_covered(self) -> bool:
        """Is the corresponding target covered."""
        return (
            len(self._solutions) == 1
            and self._capacity == 1
            and self._solutions[0].h == 1.0
        )

    # pylint: disable=invalid-name
    def add_solution(
        self, h: float, test_case_chromosome: tcc.TestCaseChromosome
    ) -> bool:
        """Add the given solution."""
        assert 0.0 <= h <= 1.0
        if h == 0.0:
            # From the MIO paper: if h = 0, the test is not added, regardless of the
            # following conditions.
            return False
        if h < 1.0 and self.is_covered:
            # candidate solution T does not cover the already fully covered target,
            # therefore there is no way it could be any better
            return False

        candidate_solution = MIOPopulationPair(h, test_case_chromosome)
        added = False

        # Does the candidate fully cover the target?
        if h == 1.0:
            # Yes. Has the target been fully covered by a previous solution?
            if self.is_covered:
                current_solution = self._solutions[0]
                if self._is_pair_better_than_current(
                    current_solution, candidate_solution
                ):
                    added = True
                    self._solutions[0] = candidate_solution
            else:
                # as the target is now fully covered by the candidate solution T, from
                # now on there is no need to keep more than one solution, only the
                # single best one. therefore, we can get rid of all solutions (if any)
                # and shrink the number of solutions to only one.
                added = True
                self._capacity = 1
                self._solutions.clear()
                self._solutions.append(candidate_solution)
        else:
            # No, candidate does not fully cover the target.
            # Is there still space for another solution?
            if len(self._solutions) < self._capacity:
                # Yes, there is
                added = True
                self._solutions.append(candidate_solution)
            else:
                worst_solution = self._solutions[-1]
                if self._is_pair_better_than_current(
                    worst_solution, candidate_solution
                ):
                    added = True
                    self._solutions[-1] = candidate_solution
            self._sort_solutions()

        assert len(self._solutions) <= self._capacity
        if added:
            self._counter = 0
        return added

    @property
    def num_solutions(self):
        """The current of number of contained solutions."""
        return len(self._solutions)

    def sample_solution(self) -> tcc.TestCaseChromosome | None:
        """Sample a random solution, if possible."""
        if len(self._solutions) == 0:
            return None
        self._counter += 1
        return randomness.choice(self._solutions).test_case_chromosome

    def get_best_solution_if_any(self) -> tcc.TestCaseChromosome | None:
        """Get the best solution, if there is one."""
        if self.is_covered:
            return self._solutions[0].test_case_chromosome
        return None

    def shrink_population(self, new_population_size) -> None:
        """Shrink the size of the population."""
        assert new_population_size > 0
        if self.is_covered:
            return
        self._capacity = new_population_size
        self._solutions = self._solutions[0:new_population_size]

    def _sort_solutions(self):
        # Desc sort, from highest to lowest h.
        self._solutions.sort(key=lambda c: c.h, reverse=True)

    @staticmethod
    def _is_pair_better_than_current(
        current: MIOPopulationPair, candidate: MIOPopulationPair
    ):
        if current.h > candidate.h:
            return False
        if current.h < candidate.h:
            return True
        return MIOPopulation._is_better_than_current(
            current.test_case_chromosome, candidate.test_case_chromosome
        )

    @staticmethod
    def _is_better_than_current(
        current: tcc.TestCaseChromosome, candidate: tcc.TestCaseChromosome
    ) -> bool:
        current_result = current.get_last_execution_result()
        candidate_result = candidate.get_last_execution_result()
        if current_result is not None and (
            current_result.timeout or current_result.has_test_exceptions()
        ):
            # If the current solution has a timeout or throws an exception then a
            # solution that does neither is considered better.
            if (
                candidate_result is not None
                and not candidate_result.timeout
                and not candidate_result.has_test_exceptions()
            ):
                return True

        # Compare length otherwise
        return candidate.size() <= current.size()
        # TODO(fk) support other secondary objectives?


class MIOArchive(Archive):
    """The archive that is used in MIO."""

    def __init__(
        self,
        targets: OrderedSet[ff.TestCaseFitnessFunction],
        initial_size: int,
    ):
        super().__init__()
        self._archive: dict[ff.TestCaseFitnessFunction, MIOPopulation] = {
            target: MIOPopulation(initial_size) for target in targets
        }

    def update(self, solutions: Iterable[tcc.TestCaseChromosome]) -> bool:
        """Update the archive with the given solutions."""
        updated = False
        for solution in solutions:
            solution = solution.clone()
            for target in self._archive:
                fitness_value = solution.get_fitness_for(target)
                result = solution.get_last_execution_result()
                assert result is not None
                if result.has_test_exceptions():
                    chop_position = solution.get_last_mutatable_statement()
                    assert chop_position is not None
                    solution.test_case.chop(chop_position)
                covered_before = self._archive[target].is_covered
                updated |= self._archive[target].add_solution(
                    1.0 - ff.normalise(fitness_value), solution
                )
                # The goal was covered with this solution
                # TODO(fk) replace with goal.is_covered?
                if not covered_before and self._archive[target].is_covered:
                    self._on_target_covered(target)
        return updated

    def get_solution(self) -> tcc.TestCaseChromosome | None:
        """Get a random solution."""

        # Choose one target at random that has not been covered but contains some
        # solutions. In case there is not any non-covered target with at least one
        # solution, either because all targets have been covered or for the non-covered
        # targets there is not any solution yet, then choose one of the covered targets
        # at random. Thereafter, choose one solution randomly from the list of solutions
        # of the chosen target.
        targets_with_solutions = [
            func
            for func, population in self._archive.items()
            if population.num_solutions > 0
        ]
        if len(targets_with_solutions) == 0:
            # There is not at least one target with at least one solution
            return None

        potential_targets = [
            func
            for func in targets_with_solutions
            if not self._archive[func].is_covered
        ]
        if len(potential_targets) == 0:
            potential_targets = targets_with_solutions

        assert len(potential_targets) > 0

        # We shuffle before, so we don't always pick the same, in case of ties.
        randomness.RNG.shuffle(potential_targets)

        # Instead of choosing a target at random, we choose the one with the lowest
        # counter value. (See Section 3.3 of the paper that describes this archive
        # for more details)
        potential_targets.sort(key=lambda t: self._archive[t].counter)
        sampled = self._archive[potential_targets[0]].sample_solution()
        if sampled is not None:
            sampled = sampled.clone()
        return sampled

    def shrink_solutions(self, new_population_size):
        """Shrink all populations to the new given size."""
        assert new_population_size > 0
        for population in self._archive.values():
            population.shrink_population(new_population_size)

    @property
    def solutions(self) -> OrderedSet[tcc.TestCaseChromosome]:
        result = OrderedSet()
        for population in self._archive.values():
            solution = population.get_best_solution_if_any()
            if solution is not None:
                result.add(solution)
        return result

    @property
    def num_covered_targets(self) -> int:
        """The amount of targets that are covered."""
        return len(
            [1 for population in self._archive.values() if population.is_covered]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pynguin',
 'pynguin.analyses',
 'pynguin.analyses.controlflow',
 'pynguin.analyses.module',
 'pynguin.analyses.seeding',
 'pynguin.analyses.seeding.testimport',
 'pynguin.assertion',
 'pynguin.assertion.mutation_analysis',
 'pynguin.coverage',
 'pynguin.ga',
 'pynguin.ga.comparators',
 'pynguin.ga.operators',
 'pynguin.ga.operators.crossover',
 'pynguin.ga.operators.ranking',
 'pynguin.ga.operators.selection',
 'pynguin.generation',
 'pynguin.generation.algorithms',
 'pynguin.generation.export',
 'pynguin.generation.stoppingconditions',
 'pynguin.instrumentation',
 'pynguin.resources',
 'pynguin.setup',
 'pynguin.testcase',
 'pynguin.typeinference',
 'pynguin.utils',
 'pynguin.utils.generic',
 'pynguin.utils.statistics']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0,<4.0',
 'MutPy-Pynguin>=0.7.1,<0.8.0',
 'Pygments>=2.11,<3.0',
 'astor>=0.8.1,<0.9.0',
 'bytecode>=0,<1',
 'jellyfish>=0,<1',
 'networkx[pydot]>=2,<3',
 'ordered-set>=4.0,<5.0',
 'pydot>=1.4,<2.0',
 'pytest>=6.2,<7.0',
 'rich>=11.0,<12.0',
 'simple-parsing>=0.0.17,<0.0.18',
 'typing_inspect>=0,<1']

entry_points = \
{'console_scripts': ['pynguin = pynguin.cli:main']}

setup_kwargs = {
    'name': 'pynguin',
    'version': '0.16.1',
    'description': 'Pynguin is a tool for automated unit test generation for Python',
    'long_description': '<!--\nSPDX-FileCopyrightText: 2019–2022 Pynguin Contributors\n\nSPDX-License-Identifier: CC-BY-4.0\n-->\n\n# Pynguin\n\nPynguin (IPA: ˈpɪŋɡuiːn),\nthe\nPYthoN\nGeneral\nUnIt\ntest\ngeNerator,\nis a tool that allows developers to generate unit tests automatically.\n\nTesting software is often considered to be a tedious task.\nThus, automated generation techniques have been proposed and mature tools exist—for\nstatically typed languages, such as Java.\nThere is, however, no fully-automated tool available that produces unit tests for\ngeneral-purpose programs in a dynamically typed language.\nPynguin is, to the best of our knowledge, the first tool that fills this gap\nand allows the automated generation of unit tests for Python programs.\n\nPynguin is developed at the\n[Chair of Software Engineering II](https://www.fim.uni-passau.de/lehrstuhl-fuer-software-engineering-ii/) \nof the [University of Passau](https://www.uni-passau.de).\n\n<details>\n<summary>Internal Pipeline Status</summary>\n\n[![pipeline status](https://gitlab.infosun.fim.uni-passau.de/se2/pynguin/pynguin/badges/main/pipeline.svg)](https://gitlab.infosun.fim.uni-passau.de/se2/pynguin/pynguin/-/commits/main)\n[![coverage report](https://gitlab.infosun.fim.uni-passau.de/se2/pynguin/pynguin/badges/main/coverage.svg)](https://gitlab.infosun.fim.uni-passau.de/se2/pynguin/pynguin/-/commits/main)  \n\n</details>\n\n[![License LGPL v3](https://img.shields.io/badge/License-LGPL%20v3-blue.svg)](https://www.gnu.org/licenses/lgpl-3.0)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![PyPI version](https://badge.fury.io/py/pynguin.svg)](https://badge.fury.io/py/pynguin)\n[![Supported Python Versions](https://img.shields.io/pypi/pyversions/pynguin.svg)](https://github.com/se2p/pynguin)\n[![Documentation Status](https://readthedocs.org/projects/pynguin/badge/?version=latest)](https://pynguin.readthedocs.io/en/latest/?badge=latest)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3989840.svg)](https://doi.org/10.5281/zenodo.3989840)\n[![REUSE status](https://api.reuse.software/badge/github.com/se2p/pynguin)](https://api.reuse.software/info/github.com/se2p/pynguin)\n[![Downloads](https://static.pepy.tech/personalized-badge/pynguin?period=total&units=international_system&left_color=grey&right_color=blue&left_text=Downloads)](https://pepy.tech/project/pynguin)\n\n\n![Pynguin Logo](https://raw.githubusercontent.com/se2p/pynguin/master/docs/source/_static/pynguin-logo.png "Pynguin Logo")\n\n## Attention\n\nPynguin executes the module under test!\nAs a consequence, depending on what code is in that module,\nrunning Pynguin can cause serious harm to your computer,\nfor example, wipe your entire hard disk!\nWe recommend running Pynguin in an isolated environment;\nuse, for example, a Docker container to minimize the risk of damaging\nyour system.\n\n\n## Prerequisites\n\nBefore you begin, ensure you have met the following requirements:\n- You have installed Python 3.8, 3.9, or 3.10 (we have not yet tested with Python\n  3.11, there might be some problems due to changed internals regarding the byte-code\n  instrumentation).\n- You have a recent Linux/macOS/Windows machine.\n\nPlease consider reading the [online documentation](https://pynguin.readthedocs.io)\nto start your Pynguin adventure.\n \n## Installing Pynguin\n\nPynguin can be easily installed using the `pip` tool by typing:\n```bash\npip install pynguin\n```\n\nMake sure that your version of `pip` is that of a supported Python version, as any \nolder version is not supported by Pynguin!\n\n## Using Pynguin\n\nBefore you continue, please read the [quick start guide](https://pynguin.readthedocs.io/en/latest/user/quickstart.html)\n\nPynguin is a command-line application.\nOnce you installed it to a virtual environment, you can invoke the tool by typing\n`pynguin` inside this virtual environment.\nPynguin will then print a list of its command-line parameters.\n\nA minimal full command line to invoke Pynguin could be the following,\nwhere we assume that a project `foo` is located in `/tmp/foo`,\nwe want to store Pynguin\'s generated tests in `/tmp/testgen`,\nand we want to generate tests using a whole-suite approach for the module `foo.bar`\n(wrapped for better readability):\n```bash\npynguin \\\n  --project-path /tmp/foo \\\n  --output-path /tmp/testgen \\\n  --module-name foo.bar\n```\nPlease find a more detailed example in the [quick start guide](https://pynguin.readthedocs.io/en/latest/user/quickstart.html).\n\n\n## Contributing to Pynguin\n\nFor the development of Pynguin you will need the [`poetry`](https://python-poetry.org)\ndependency management and packaging tool.\nTo start developing, follow these steps:\n1. Clone the repository\n2. Change to the `pynguin` folder: `cd pynguin`\n3. Create a virtual environment and install dependencies using `poetry`: `poetry install`\n4. Make your changes\n5. Run `poetry shell` to switch to the virtual environment in your current shell\n6. Run `make check` to verify that your changes pass all checks\n\n   Please see the [`poetry` documentation](https://python-poetry.org/docs/) for more information on this tool.\n\n### Development using PyCharm.\n\nIf you want to use the PyCharm IDE you have to set up a few things:\n1. Import pynguin into PyCharm.\n2. Let the poetry support configure a virtual environment.\n\n   If you use a PyCharm version before `2021.3`, you have to install the poetry support as a [plugin](https://plugins.jetbrains.com/plugin/14307-poetry).\n3. Set the default test runner to `pytest`\n4. Set the DocString format to `Google`\n\n\n## License\n\nThis project is licensed under the terms of the\n[GNU Lesser General Public License](LICENSE.rst).\n',
    'author': 'Stephan Lukasczyk',
    'author_email': 'stephan@lukasczyk.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.pynguin.eu',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
